#
# Copyright (c) 2002, 2003 Art Haas
#
# This file is part of PythonCAD.
#
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# code to calculate if or where two objects intersect
#

# from __future__ import division

import point
import segment
import circle
import arc
import hcline
import vcline
import acline
import cline
import ccircle
import math
import warnings

#
# common constants
#

_dtr = math.pi/180.0
_rtd = 180.0/math.pi

#
# the following functions are used to calculate the
# intersection of two line segments
#
# see comp.graphics.algorithms FAQ for details
#

def denom(p1, p2, p3, p4):
    if not isinstance(p1, point.Point):
        raise TypeError, "Invalid argument to denom(): " + `p1`
    if not isinstance(p2, point.Point):
        raise TypeError, "Invalid argument to denom(): " + `p2`
    if not isinstance(p3, point.Point):
        raise TypeError, "Invalid argument to denom(): " + `p3`
    if not isinstance(p4, point.Point):
        raise TypeError, "Invalid argument to denom(): " + `p4`
    _p1x, _p1y = p1.getCoords()
    _p2x, _p2y = p2.getCoords()
    _p3x, _p3y = p3.getCoords()
    _p4x, _p4y = p4.getCoords()
    return ((_p2x - _p1x)*(_p4y - _p3y)) - ((_p2y - _p1y)*(_p4x - _p3x))

def rnum(p1, p2, p3, p4):
    if not isinstance(p1, point.Point):
        raise TypeError, "Invalid argument to rnum(): " + `p1`
    if not isinstance(p2, point.Point):
        raise TypeError, "Invalid argument to rnum(): " + `p2`
    if not isinstance(p3, point.Point):
        raise TypeError, "Invalid argument to rnum(): " + `p3`
    if not isinstance(p4, point.Point):
        raise TypeError, "Invalid argument to rnum(): " + `p4`
    _p1x, _p1y = p1.getCoords()
    _p2x, _p2y = p2.getCoords()
    _p3x, _p3y = p3.getCoords()
    _p4x, _p4y = p4.getCoords()
    return ((_p1y - _p3y)*(_p4x - _p3x)) - ((_p1x - _p3x)*(_p4y - _p3y))

def snum(p1, p2, p3, p4):
    if not isinstance(p1, point.Point):
        raise TypeError, "Invalid argument to snum(): " + `p1`
    if not isinstance(p2, point.Point):
        raise TypeError, "Invalid argument to snum(): " + `p2`
    if not isinstance(p3, point.Point):
        raise TypeError, "Invalid argument to snum(): " + `p3`
    if not isinstance(p4, point.Point):
        raise TypeError, "Invalid argument to snum(): " + `p4`
    _p1x, _p1y = p1.getCoords()
    _p2x, _p2y = p2.getCoords()
    _p3x, _p3y = p3.getCoords()
    _p4x, _p4y = p4.getCoords()
    return ((_p1y - _p3y)*(_p2x - _p1x)) - ((_p1x - _p3x)*(_p2y - _p1y))

#
# the following function is used to calculate the intersection
# of a circle and a segment or a circle and an infite length line
#
# see Paul Bourke's web pages for details
#

def sc_intersection(x1, y1, x2, y2 ,_circ, skiptest=False):
    if not isinstance(x1, float):
        raise TypeError, "Invalid argument: " + `x1`
    if not isinstance(y1, float):
        raise TypeError, "Invalid argument: " + `y1`
    if not isinstance(x2, float):
        raise TypeError, "Invalid argument: " + `x2`
    if not isinstance(y2, float):
        raise TypeError, "Invalid argument: " + `y2`
    if not isinstance(_circ, (circle.Circle,
                              arc.Arc,
                              ccircle.CCircle)):
        raise TypeError, "Invalid circle for sc_intersection(): " + `_circ`
    _plist = []
    _xdiff = x2 - x1
    _ydiff = y2 - y1
    _xc, _yc = _circ.getCenter().getCoords()
    _r = _circ.getRadius()
    _u = ((_xc - x1)*(_xdiff) + (_yc - y1)*(_ydiff))/(pow(_xdiff, 2) + pow(_ydiff, 2))
    _xp = x1 + (_u * _xdiff)
    _yp = y1 + (_u * _ydiff)
    _dist = math.hypot((_yp - _yc), (_xp - _xc))
    if abs(_dist - _r) < 1e-10:
        if skiptest or not (_u < 0.0 or _u > 1.0):
            _plist.append((_xp, _yp))
    else:
        if _dist < _r:
            _a = pow(_xdiff, 2) + pow(_ydiff, 2)
            _b = 2.0 * ((_xdiff * (x1 - _xc)) + (_ydiff * (y1 - _yc)))
            _c = (pow(_xc, 2) + pow(_yc, 2) + pow(x1, 2) + pow(y1, 2) -
                  2.0 * ((_xc*x1) + (_yc*y1)) - pow(_r,2))
            _det = pow(_b, 2) - (4.0 * _a * _c)
            if _det > 1e-10:
                _dsq = math.sqrt(_det)
                _u = (-_b - _dsq)/(2.0 * _a)
                if skiptest or not (_u < 0.0 or _u > 1.0):
                    _x = x1 + (_u * _xdiff)
                    _y = y1 + (_u * _ydiff)
                    _plist.append((_x, _y))
                _u = (-_b + _dsq)/(2.0 * _a)
                if skiptest or not (_u < 0.0 or _u > 1.0):
                    _x = x1 + (_u * _xdiff)
                    _y = y1 + (_u * _ydiff)
                    _plist.append((_x, _y))
    return _plist

#
# intersection functions
#

def _null_intfunc(ipts, obja, objb):
    print "invoked _null_intfunc()"
    print "obja: " + `obja`
    print "objb: " + `objb`

def _non_intersecting(ipts, obja, objb):
    pass

#
# segment - segment intersection
#

def _seg_seg_intersection(ipts, seg1, seg2):
    _p1, _p2 = seg1.getEndpoints()
    _p3, _p4 = seg2.getEndpoints()
    _xi = _yi = None
    if _p1 == _p3:
        _x, _y = _p4.getCoords()
        _proj = seg1.getProjection(_x, _y)
        if _proj is None:
            _xi, _yi = _p1.getCoords()
        else:
            _px, _py = _proj
            if math.hypot((_px - _x), (_py - _y)) > 1e-10:
                _xi, _yi = _p1.getCoords()
    elif _p1 == _p4:
        _x, _y = _p3.getCoords()
        _proj = seg1.getProjection(_x, _y)
        if _proj is None:
            _xi, _yi = _p1.getCoords()
        else:
            _px, _py = _proj
            if math.hypot((_px - _x), (_py - _y)) > 1e-10:
                _xi, _yi = _p1.getCoords()
    elif _p2 == _p3:
        _x, _y = _p4.getCoords()
        _proj = seg1.getProjection(_x, _y)
        if _proj is None:
            _xi, _yi = _p2.getCoords()
        else:
            _px, _py = _proj
            if math.hypot((_px - _x), (_py - _y)) > 1e-10:
                _xi, _yi = _p2.getCoords()
    elif _p2 == _p4:
        _x, _y = _p3.getCoords()
        _proj = seg1.getProjection(_x, _y)
        if _proj is None:
            _xi, _yi = _p2.getCoords()
        else:
            _px, _py = _proj
            if math.hypot((_px - _x), (_py - _y)) > 1e-10:
                _xi, _yi = _p2.getCoords()
    else:
        _d = denom(_p1, _p2, _p3, _p4)
        if abs(_d) > 1e-10: # NOT parallel
            _rn = rnum(_p1, _p2, _p3, _p4)
            if abs(_rn) > 1e-10: # NOT colinear
                _sn = snum(_p1, _p2, _p3 ,_p4)
                _r = _rn/_d
                _s = _sn/_d
                if 0.0 < _r < 1.0 and 0.0 < _s < 1.0:
                    _p1x, _p1y = _p1.getCoords()
                    _p2x, _p2y = _p2.getCoords()
                    _xi = _p1x + _r * (_p2x - _p1x)
                    _yi = _p1y + _r * (_p2y - _p1y)
    if _xi is not None and _yi is not None:
        ipts.append((_xi, _yi))

#
# segment - circle intersection
#

def _seg_circ_intersection(ipts, obja, objb):
    if isinstance(obja, segment.Segment):
        _seg = obja
        _circ = objb
    else:
        _seg = objb
        _circ = obja
    assert not isinstance(_circ, arc.Arc), "Arc found!"
    _p1, _p2 = _seg.getEndpoints()
    _p1x, _p1y = _p1.getCoords()
    _p2x, _p2y = _p2.getCoords()
    for _ip in sc_intersection(_p1x, _p1y, _p2x, _p2y, _circ):
        ipts.append(_ip)

#
# segment - arc intersection
#

def _seg_arc_intersection(ipts, obja, objb):
    if isinstance(obja, segment.Segment):
        _seg = obja
        _arc = objb
    else:
        _seg = objb
        _arc = obja
    _p1, _p2 = _seg.getEndpoints()
    _p1x, _p1y = _p1.getCoords()
    _p2x, _p2y = _p2.getCoords()
    _ax, _ay = _arc.getCenter().getCoords()
    for _ip in sc_intersection(_p1x, _p1y, _p2x, _p2y, _arc):
        _x, _y = _ip
        _angle = math.atan2((_y - _ay),(_x - _ax)) * _rtd
        if _angle < 0.0:
            _angle = _angle + 360.0
        if _arc.throughAngle(_angle):
            ipts.append(_ip)

#
# segment - horizontal construction line intersection
#

def _seg_hcl_intersection(ipts, obja, objb):
    if isinstance(obja, segment.Segment):
        _seg = obja
        _hcl = objb
    else:
        _seg = objb
        _hcl = obja
    _p1, _p2 = _seg.getEndpoints()
    _hp = _hcl.getLocation()
    _hx, _hy = _hp.getCoords()
    _xi = _yi = None
    if _hp == _p1:
        _p2x, _p2y = _p2.getCoords()
        if abs(_hy - _p2y) > 1e-10:
            _xi, _yi = _p1.getCoords()
    elif _hp == _p2:
        _p1x, _p1y = _p1.getCoords()
        if abs(_hy - _p1y) > 1e-10:
            _xi, _yi = _p2.getCoords()
    else:
        _p1x, _p1y = _p1.getCoords()
        _p2x, _p2y = _p2.getCoords()
        _miny = min(_p1y, _p2y)
        _maxy = max(_p1y, _p2y)
        if abs(_p1y - _p2y) > 1e-10 and _miny < _hy < _maxy:
            _xi = _p1x + ((_hy - _p1y)*(_p2x - _p1x))/(_p2y - _p1y)
            _yi = _hy
    if _xi is not None and _yi is not None:
        ipts.append((_xi, _yi))

#
# segment - vertical construction line intersection
#

def _seg_vcl_intersection(ipts, obja, objb):
    if isinstance(obja, segment.Segment):
        _seg = obja
        _vcl = objb
    else:
        _seg = objb
        _vcl = obja
    _p1, _p2 = _seg.getEndpoints()
    _vp = _vcl.getLocation()
    _vx, _vy = _vp.getCoords()
    _xi = _yi = None
    if _vp == _p1:
        _p2x, _p2y = _p2.getCoords()
        if abs(_vx - _p2x) > 1e-10:
            _xi, _yi = _p1.getCoords()
    elif _vp == _p2:
        _p1x, _p1y = _p1.getCoords()
        if abs(_vx - _p1x) > 1e-10:
            _xi, _yi = _p2.getCoords()
    else:
        _p1x, _p1y = _p1.getCoords()
        _p2x, _p2y = _p2.getCoords()
        _vx, _vy = _vp.getCoords()
        _minx = min(_p1x, _p2x)
        _maxx = max(_p1x, _p2x)
        if abs(_p1x - _p2x) > 1e-10 and _minx < _vx < _maxx:
            _xi = _vx
            _yi = _p1y + ((_p2y - _p1y)*(_vx - _p1x))/(_p2x - _p1x)
    if _xi is not None and _yi is not None:
        ipts.append((_xi, _yi))

#
# segment - angled construction line intersection
#

def _seg_acl_intersection(ipts, obja, objb):
    if isinstance(obja, segment.Segment):
        _seg = obja
        _acl = objb
    else:
        _seg = objb
        _acl = obja
    _p1, _p2 = _seg.getEndpoints()
    _ap = _acl.getLocation()
    _xi = _yi = None
    if _p1 == _ap:
        _x, _y = _p2.getCoords()
        _px, _py = _acl.getProjection(_x, _y)
        if math.hypot((_px - _x), (_py - _y)) > 1e-10:
            _xi, _yi = _p1.getCoords()
    elif _p2 == _ap:
        _x, _y = _p1.getCoords()
        _px, _py = _acl.getProjection(_x, _y)
        if math.hypot((_px - _x), (_py - _y)) > 1e-10:
            _xi, _yi = _p2.getCoords()
    else:
        _ax, _ay = _ap.getCoords()
        _angle = _acl.getAngle() * _dtr
        _xt = _ax + math.cos(_angle)
        _yt = _ay + math.sin(_angle)
        _tp = point.Point(_xt, _yt)
        _d = denom(_p1, _p2, _ap, _tp)
        if abs(_d) > 1e-10: # NOT parallel
            _rn = rnum(_p1, _p2, _ap, _tp)
            if abs(_rn) > 1e-10: # NOT colinear
                _r = _rn/_d
                if 0.0 < _r < 1.0:
                    _p1x, _p1y = _p1.getCoords()
                    _p2x, _p2y = _p2.getCoords()
                    _xi = _p1x + _r * (_p2x - _p1x)
                    _yi = _p1y + _r * (_p2y - _p1y)
    if _xi is not None and _yi is not None:
        ipts.append((_xi, _yi))

#
# segment - 2-point construction line intersection
#

def _seg_cl_intersection(ipts, obja, objb):
    if isinstance(obja, segment.Segment):
        _seg = obja
        _cl = objb
    else:
        _seg = objb
        _cl = obja
    _p1, _p2 = _seg.getEndpoints()
    _p3, _p4 = _cl.getKeypoints()
    _xi = _yi = None
    if _p1 == _p3 or _p1 == _p4:
        _x, _y = _p2.getCoords()
        _px, _py = _cl.getProjection(_x, _y)
        if math.hypot((_px - _x), (_py - _y)) > 1e-10:
            _xi, _yi = _p1.getCoords()
    elif _p2 == _p3 or _p2 == _p4:
        _x, _y = _p1.getCoords()
        _px, _py = _cl.getProjection(_x, _y)
        if math.hypot((_px - _x), (_py - _y)) > 1e-10:
            _xi, _yi = _p2.getCoords()
    else:
        _d = denom(_p1, _p2, _p3, _p4)
        if abs(_d) > 1e-10: # NOT parallel
            _rn = rnum(_p1, _p2, _p3, _p4)
            if abs(_rn) > 1e-10: # NOT colinear
                _r = _rn/_d
                if 0.0 < _r < 1.0:
                    _p1x, _p1y = _p1.getCoords()
                    _p2x, _p2y = _p2.getCoords()
                    _xi = _p1x + _r * (_p2x - _p1x)
                    _yi = _p1y + _r * (_p2y - _p1y)
    if _xi is not None and _yi is not None:
        ipts.append((_xi, _yi))

#
# circle - circle intersection
#
# see Paul Bourke's web pages for algorithm
#

def _circ_circ_intersection(ipts, circ1, circ2):
    assert not isinstance(circ1, arc.Arc), "Arc found!"
    assert not isinstance(circ2, arc.Arc), "Arc found!"
    _c1x, _c1y = circ1.getCenter().getCoords()
    _r1 = circ1.getRadius()
    _c2x, _c2y = circ2.getCenter().getCoords()
    _r2 = circ2.getRadius()
    _xdiff = _c2x - _c1x
    _ydiff = _c2y - _c1y
    if not (abs(_xdiff) < 1e-10 and abs(_ydiff) < 1e-10): # NOT concentric
        _sep = math.hypot(_xdiff, _ydiff)
        _maxsep = _r1 + _r2 + 1e-10
        _minsep = abs(_r1 - _r2) - 1e-10
        # print "sep: %g; maxsep: %g; minsep: %g" % (_sep, _maxsep, _minsep)
        if _minsep < _sep < _maxsep:
            _a = (pow(_r1, 2) - pow(_r2, 2) + pow(_sep, 2))/(2*_sep)
            _pcx = _c1x + _a * (_xdiff/_sep)
            _pcy = _c1y + _a * (_ydiff/_sep)
            # print "a: %g; pcx: %g; pcy: %g" % (_a, _pcx, _pcy)
            if abs(abs(_a) - _r1) < 1e-10: # single contact point
                ipts.append((_pcx, _pcy))
            else: # two contact points
                _h = math.sqrt(pow(_r1, 2) - pow(_a, 2))
                _x = _pcx + _h * (_ydiff/_sep)
                _y = _pcy - _h * (_xdiff/_sep)
                ipts.append((_x, _y))
                _x = _pcx - _h * (_ydiff/_sep)
                _y = _pcy + _h * (_xdiff/_sep)
                ipts.append((_x, _y))

#
# circle - arc intersection
#

def _circ_arc_intersection(ipts, obja, objb):
    if isinstance(obja, arc.Arc):
        _arc = obja
        _circ = objb
    else:
        _circ = obja
        _arc = objb
    _cp = _arc.getCenter()
    _r = _arc.getRadius()
    _tempcirc = circle.Circle(_cp, _r)
    _ipts = []
    _circ_circ_intersection(_ipts, _circ, _tempcirc)
    if len(_ipts): # may have intersection points ...
        _cx, _cy = _cp.getCoords()
        for _ip in _ipts:
            _x, _y = _ip
            _angle = math.atan2((_y - _cy),(_x - _cx)) * _rtd
            if _angle < 0.0:
                _angle = _angle + 360.0
            if _arc.throughAngle(_angle):
                ipts.append(_ip)

#
# circle - horizontal contruction line intersection
#

def _circ_hcl_intersection(ipts, obja, objb):
    if isinstance(obja, (circle.Circle, ccircle.CCircle)):
        _circ = obja
        _hcl = objb
    else:
        _circ = objb
        _hcl = obja
    _cp = _circ.getCenter()
    _r = _circ.getRadius()
    _hp = _hcl.getLocation()
    _hx, _hy = _hp.getCoords()
    if _hp == _cp:
        ipts.append(((_hx - _r), _hy))
        ipts.append(((_hx + _r), _hy))
    else:
        _cx, _cy = _circ.getCenter().getCoords()
        _ymin = _cy - _r
        _ymax = _cy + _r
        if abs(_ymin - _hy) < 1e-10:
            ipts.append((_cx, _ymin))
        elif abs(_ymax - _hy) < 1e-10:
            ipts.append((_cx, _ymax))
        elif _ymin < _hy < _ymax:
            _offset = math.sqrt(pow(_r, 2) - pow((_cy - _hy), 2))
            ipts.append(((_cx - _offset), _hy))
            ipts.append(((_cx + _offset), _hy))
        else:
            pass
#
# circle - vertical contruction line intersection
#

def _circ_vcl_intersection(ipts, obja, objb):
    if isinstance(obja, (circle.Circle, ccircle.CCircle)):
        _circ = obja
        _vcl = objb
    else:
        _circ = objb
        _vcl = obja
    _cp = _circ.getCenter()
    _r = _circ.getRadius()
    _vp = _vcl.getLocation()
    _vx, _vy = _vp.getCoords()
    if _vp == _cp:
        ipts.append((_vx, (_vy - _r)))
        ipts.append((_vx, (_vy + _r)))
    else:
        _cx, _cy = _circ.getCenter().getCoords()
        _xmin = _cx - _r
        _xmax = _cx + _r
        if abs(_xmin - _vx) < 1e-10:
            ipts.append((_xmin, _cy))
        elif abs(_xmax - _vx) < 1e-10:
            ipts.append((_xmax, _cy))
        elif _xmin < _vx < _xmax:
            _offset = math.sqrt(pow(_r, 2) - pow((_vx - _cx), 2))
            ipts.append((_vx, (_cy - _offset)))
            ipts.append((_vx, (_cy + _offset)))
        else:
            pass

#
# circle - angled construction line intersection
#

def _circ_acl_intersection(ipts, obja, objb):
    if isinstance(obja, (circle.Circle, ccircle.CCircle)):
        _circ = obja
        _acl = objb
    else:
        _circ = objb
        _acl = obja
    assert not isinstance(_circ, arc.Arc), "Arc found!"
    _p1x, _p1y = _acl.getLocation().getCoords()
    _angle = _acl.getAngle() * _dtr
    _xt = _p1x + math.cos(_angle)
    _yt = _p1y + math.sin(_angle)
    for _ip in sc_intersection(_p1x, _p1y, _xt, _yt, _circ, True):
        ipts.append(_ip)

#
# circle - 2-point construction line intersection
#

def _circ_cl_intersection(ipts, obja, objb):
    if isinstance(obja, (circle.Circle, ccircle.CCircle)):
        _circ = obja
        _cl = objb
    else:
        _circ = objb
        _cl = obja
    assert not isinstance(_circ, arc.Arc), "Arc found!"
    _p1, _p2 = _cl.getKeypoints()
    _p1x, _p1y = _p1.getCoords()
    _p2x, _p2y = _p2.getCoords()
    for _ip in sc_intersection(_p1x, _p1y, _p2x, _p2y, _circ, True):
        ipts.append(_ip)

#
# arc - arc intersection
#

def _arc_arc_intersection(ipts, arc1, arc2):
    _cp1 = arc1.getCenter()
    _r1 = arc1.getRadius()
    _tempcirc1 = circle.Circle(_cp1, _r1)
    _cp2 = arc2.getCenter()
    _r2 = arc2.getRadius()
    _tempcirc2 = circle.Circle(_cp2, _r2)
    _ipts = []
    _circ_circ_intersection(_ipts, _tempcirc1, _tempcirc2)
    if len(_ipts): # may have intersection points ...
        for _ip in _ipts:
            _cx, _cy = _cp1.getCoords()
            _x, _y = _ip
            _angle = math.atan2((_y - _cy), (_x - _cx)) * _rtd
            if _angle < 0.0:
                _angle = _angle + 360.0
            if arc1.throughAngle(_angle):
                _cx, _cy = _cp2.getCoords()
                _angle = math.atan2((_y - _cy), (_x - _cx)) * _rtd
                if _angle < 0.0:
                    _angle = _angle + 360.0
                if arc2.throughAngle(_angle):
                    ipts.append(_ip)

#
# arc - horizontal construction line intersection
#

def _arc_hcl_intersection(ipts, obja, objb):
    if isinstance(obja, arc.Arc):
        _arc = obja
        _hcl = objb
    else:
        _arc = objb
        _hcl = obja
    _cp = _arc.getCenter()
    _r = _arc.getRadius()
    _tempcirc = circle.Circle(_cp, _r)
    _ipts = []
    _circ_hcl_intersection(_ipts, _tempcirc, _hcl)
    if len(_ipts): # may have intersection points ...
        _cx, _cy = _cp.getCoords()
        for _ip in _ipts:
            _x, _y = _ip
            _angle = math.atan2((_y - _cy),(_x - _cx)) * _rtd
            if _angle < 0.0:
                _angle = _angle + 360.0
            if _arc.throughAngle(_angle):
                ipts.append(_ip)

#
# arc - vertical construction line intersection
#

def _arc_vcl_intersection(ipts, obja, objb):
    if isinstance(obja, arc.Arc):
        _arc = obja
        _vcl = objb
    else:
        _arc = objb
        _vcl = obja
    _cp = _arc.getCenter()
    _r = _arc.getRadius()
    _tempcirc = circle.Circle(_cp, _r)
    _ipts = []
    _circ_vcl_intersection(_ipts, _tempcirc, _vcl)
    if len(_ipts): # may have intersection points ...
        _cx, _cy = _cp.getCoords()
        for _ip in _ipts:
            _x, _y = _ip
            _angle = math.atan2((_y - _cy),(_x - _cx)) * _rtd
            if _angle < 0.0:
                _angle = _angle + 360.0
            if _arc.throughAngle(_angle):
                ipts.append(_ip)

#
# arc - angled construction line intersection
#

def _arc_acl_intersection(ipts, obja, objb):
    if isinstance(obja, arc.Arc):
        _arc = obja
        _acl = objb
    else:
        _arc = objb
        _acl = obja
    _cp = _arc.getCenter()
    _r = _arc.getRadius()
    _tempcirc = circle.Circle(_cp, _r)
    _ipts = []
    _circ_acl_intersection(_ipts, _tempcirc, _acl)
    if len(_ipts): # may have intersection points ...
        _cx, _cy = _cp.getCoords()
        for _ip in _ipts:
            _x, _y = _ip
            _angle = math.atan2((_y - _cy),(_x - _cx)) * _rtd
            if _angle < 0.0:
                _angle = _angle + 360.0
            if _arc.throughAngle(_angle):
                ipts.append(_ip)

#
# arc - 2-point construction line intersection
#

def _arc_cl_intersection(ipts, obja, objb):
    if isinstance(obja, arc.Arc):
        _arc = obja
        _cl = objb
    else:
        _arc = objb
        _cl = obja
    _cp = _arc.getCenter()
    _r = _arc.getRadius()
    _tempcirc = circle.Circle(_cp, _r)
    _ipts = []
    _circ_cl_intersection(_ipts, _tempcirc, _cl)
    if len(_ipts): # may have intersection points ...
        _cx, _cy = _cp.getCoords()
        for _ip in _ipts:
            _x, _y = _ip
            _angle = math.atan2((_y - _cy),(_x - _cx)) * _rtd
            if _angle < 0.0:
                _angle = _angle + 360.0
            if _arc.throughAngle(_angle):
                ipts.append(_ip)

#
# horizontal construction line - vertical construction line intersection
#

def _hcl_vcl_intersection(ipts, obja, objb):
    if isinstance(obja, hcline.HCLine):
        _hcl = obja
        _vcl = objb
    else:
        _hcl = objb
        _vcl = obja
    _hx, _hy = _hcl.getLocation().getCoords()
    _vx, _vh = _vcl.getLocation().getCoords()
    ipts.append((_vx, _hy))

#
# horizontal construction line - angled construction line intersection
#

def _hcl_acl_intersection(ipts, obja, objb):
    if isinstance(obja, hcline.HCLine):
        _hcl = obja
        _acl = objb
    else:
        _hcl = objb
        _acl = obja
    _angle = _acl.getAngle()
    if abs(_angle) > 1e-10: # acl is NOT horizontal
        _xi = _yi = None
        _hp = _hcl.getLocation()
        _ap = _acl.getLocation()
        if _hp == _ap:
            _xi, _yi = _hp.getCoords()
        else:
            _hx, _hy = _hp.getCoords()
            _yi = _hy
            _ax, _ay = _ap.getCoords()
            if abs(abs(_angle) - 90.0) < 1e-10: # acl is vertical
                _xi = _ax
            else:
                _slope = math.tan(_angle * _dtr)
                _yint = _ay - _slope*_ax
                _xi = (_hy - _yint)/_slope
        ipts.append((_xi, _yi))

#
# horizontal construction line - 2-point construction line intersection
#

def _hcl_cl_intersection(ipts, obja, objb):
    if isinstance(obja, hcline.HCLine):
        _hcl = obja
        _cl = objb
    else:
        _hcl = objb
        _cl = obja
    _p1, _p2 = _cl.getKeypoints()
    _p1x, _p1y = _p1.getCoords()
    _p2x, _p2y = _p2.getCoords()
    _ydiff = _p2y - _p1y
    if abs(_ydiff) > 1e-10: # cline is NOT horizontal
        _xi = _yi = None
        _hp = _hcl.getLocation()
        if _hp == _p1:
            _xi, _yi = _p1.getCoords()
        elif _hp == _p2:
            _xi, _yi = _p2.getCoords()
        else:
            _hx, _hy = _hp.getCoords()
            _yi = _hy
            _xdiff = _p2x - _p1x
            if abs(_xdiff) < 1e-10: # cline is vertical
                _xi = _p1x
            else:
                _slope = _ydiff/_xdiff
                _yint = _p1y - _slope*_p1x
                _xi = (_hy - _yint)/_slope
        ipts.append((_xi, _yi))

#
# vertical construction line - angled construction line intersection
#

def _vcl_acl_intersection(ipts, obja, objb):
    if isinstance(obja, vcline.VCLine):
        _vcl = obja
        _acl = objb
    else:
        _vcl = objb
        _acl = obja
    _angle = _acl.getAngle()
    if abs(abs(_angle) - 90.0) > 1e-10: # acl is NOT vertical

        _vp = _vcl.getLocation()
        _ap = _acl.getLocation()
        if _vp == _ap:
            _xi, _yi = _vp.getCoords()
        else:
            _vx, _vy = _vp.getCoords()
            _xi = _vx
            _ax, _ay = _ap.getCoords()
            _slope = math.tan(_angle * _dtr)
            _yint = _ay - _slope*_ax
            _yi = _slope*_vx + _yint
        ipts.append((_xi, _yi))

#
# vertical construction line - 2-point construction line intersection
#

def _vcl_cl_intersection(ipts, obja, objb):
    if isinstance(obja, vcline.VCLine):
        _vcl = obja
        _cl = objb
    else:
        _vcl = objb
        _cl = obja
    _p1, _p2 = _cl.getKeypoints()
    _p1x, _p1y = _p1.getCoords()
    _p2x, _p2y = _p2.getCoords()
    _xdiff = _p2x - _p1x
    if abs(_xdiff) > 1e-10: # cline is NOT vertical
        _xi = _yi = None
        _vp = _vcl.getLocation()
        if _vp == _p1:
            _xi, _yi = _p1.getCoords()
        elif _vp == _p2:
            _xi, _yi = _p2.getCoords()
        else:
            _vx, _vy = _vp.getCoords()
            _xi = _vx
            _ydiff = _p2y - _p1y
            if abs(_ydiff) < 1e-10: # cline is horizontal
                _yi = _p1y
            else:
                _slope = _ydiff/_xdiff
                _yint = _p1y - _slope*_p1x
                _yi = _slope*_vx + _yint
        ipts.append((_xi, _yi))

#
# angled construction line - angled construction line intersection
#

def _acl_acl_intersection(ipts, acl1, acl2):
    _ap1 = acl1.getLocation()
    _ap1x, _ap1y = _ap1.getCoords()
    _angle1 = acl1.getAngle() * _dtr
    _xt1 = _ap1x + math.cos(_angle1)
    _yt1 = _ap1y + math.sin(_angle1)
    _t1 = point.Point(_xt1, _yt1)
    _ap2 = acl2.getLocation()
    _ap2x, _ap2y = _ap2.getCoords()
    _angle2 = acl2.getAngle() * _dtr
    _xt2 = _ap2x + math.cos(_angle2)
    _yt2 = _ap2y + math.sin(_angle2)
    _t2 = point.Point(_xt2, _yt2)
    _d = denom(_ap1, _t1, _ap2, _t2)
    if abs(_d) > 1e-10: # NOT parallel
        _xi = _yi = None
        if _ap1 == _ap2:
            _xi, _yi = _ap1.getCoords()
        else:
            _rn = rnum(_ap1, _t1, _ap2, _t2)
            if abs(_rn) > 1e-10: # NOT colinear
                _r = _rn/_d
                _xi = _ap1x + _r * (_xt1 - _ap1x)
                _yi = _ap1y + _r * (_yt1 - _ap1y)
        if _xi is not None and _yi is not None:
            ipts.append((_xi, _yi))

#
# angled construction line - 2-point construction line intersection
#

def _acl_cl_intersection(ipts, obja, objb):
    if isinstance(obja, acline.ACLine):
        _acl = obja
        _cl = objb
    else:
        _acl = objb
        _cl = obja
    _ap = _acl.getLocation()
    _apx, _apy = _ap.getCoords()
    _angle = _acl.getAngle() * _dtr
    _xt = _apx + math.cos(_angle)
    _yt = _apy + math.sin(_angle)
    _t1 = point.Point(_xt, _yt)
    _p1, _p2 = _cl.getKeypoints()
    _d = denom(_ap, _t1, _p1, _p2)
    if abs(_d) > 1e-10: # NOT parallel
        _xi = _yi = None
        if _ap == _p1:
            _xi, _yi = _p1.getCoords()
        elif _ap == _p2:
            _xi, _yi = _p2.getCoords()
        else:
            _rn = rnum(_ap, _t1, _p1, _p2)
            if abs(_rn) > 1e-10: # NOT colinear
                _r = _rn/_d
                _xi = _apx + _r * (_xt - _apx)
                _yi = _apy + _r * (_yt - _apy)
        if _xi is not None and _yi is not None:
            ipts.append((_xi, _yi))

#
# 2-point construction line - 2-point construction line intersection
#

def _cl_cl_intersection(ipts, cl1, cl2):
    _p1, _p2 = cl1.getKeypoints()
    _p3, _p4 = cl2.getKeypoints()
    _d = denom(_p1, _p2, _p3, _p4)
    if abs(_d) > 1e-10: # NOT parallel
        _xi = _yi = None
        if _p1 == _p3 or _p2 == _p3:
            _xi, _yi = _p3.getCoords()
        elif _p1 == _p4 or _p2 == _p4:
            _xi, _yi = _p4.getCoords()
        else:
            _rn = rnum(_p1, _p2, _p3, _p4)
            if abs(_rn) > 1e-10: # NOT colinear
                _r = _rn/_d
                _p1x, _p1y = _p1.getCoords()
                _p2x, _p2y = _p2.getCoords()
                _xi = _p1x + _r * (_p2x - _p1x)
                _yi = _p1y + _r * (_p2y - _p1y)
        if _xi is not None and _yi is not None:
            ipts.append((_xi, _yi))
#
# set segment intersecting function
#

def _segment_intfuncs(objb):
    if isinstance(objb, segment.Segment):
        _func = _seg_seg_intersection
    elif isinstance(objb, arc.Arc):
        _func = _seg_arc_intersection
    elif isinstance(objb, (circle.Circle, ccircle.CCircle)):
        _func = _seg_circ_intersection
    elif isinstance(objb, hcline.HCLine):
        _func = _seg_hcl_intersection
    elif isinstance(objb, vcline.VCLine):
        _func = _seg_vcl_intersection
    elif isinstance(objb, acline.ACLine):
        _func = _seg_acl_intersection
    elif isinstance(objb, cline.CLine):
        _func = _seg_cl_intersection
    else:
        _func = _null_intfunc
    return _func

#
# set circle intersecting function
#

def _circ_intfuncs(objb):
    if isinstance(objb, segment.Segment):
        _func = _seg_circ_intersection
    elif isinstance(objb, arc.Arc):
        _func = _circ_arc_intersection
    elif isinstance(objb, (circle.Circle, ccircle.CCircle)):
        _func = _circ_circ_intersection
    elif isinstance(objb, hcline.HCLine):
        _func = _circ_hcl_intersection
    elif isinstance(objb, vcline.VCLine):
        _func = _circ_vcl_intersection
    elif isinstance(objb, acline.ACLine):
        _func = _circ_acl_intersection
    elif isinstance(objb, cline.CLine):
        _func = _circ_cl_intersection
    else:
        _func = _null_intfunc
    return _func

#
# set arc intersecting function
#

def _arc_intfuncs(objb):
    if isinstance(objb, segment.Segment):
        _func  = _seg_arc_intersection
    elif isinstance(objb, arc.Arc):
        _func = _arc_arc_intersection
    elif isinstance(objb, (circle.Circle, ccircle.CCircle)):
        _func = _circ_arc_intersection
    elif isinstance(objb, hcline.HCLine):
        _func = _arc_hcl_intersection
    elif isinstance(objb, vcline.VCLine):
        _func = _arc_vcl_intersection
    elif isinstance(objb, acline.ACLine):
        _func = _arc_acl_intersection
    elif isinstance(objb, cline.CLine):
        _func = _arc_cl_intersection
    else:
        _func = _null_intfunc
    return _func

#
# set hcline intersecting function
#

def _hcline_intfuncs(objb):
    if isinstance(objb, segment.Segment):
        _func  = _seg_hcl_intersection
    elif isinstance(objb, arc.Arc):
        _func = _arc_hcl_intersection
    elif isinstance(objb, (circle.Circle, ccircle.CCircle)):
        _func = _circ_hcl_intersection
    elif isinstance(objb, hcline.HCLine):
        _func = _non_intersecting
    elif isinstance(objb, vcline.VCLine):
        _func =_hcl_vcl_intersection
    elif isinstance(objb, acline.ACLine):
        _func = _hcl_acl_intersection
    elif isinstance(objb, cline.CLine):
        _func = _hcl_cl_intersection
    else:
        _func = _null_intfunc
    return _func

#
# set vcline intersecting function
#

def _vcline_intfuncs(objb):
    if isinstance(objb, segment.Segment):
        _func  = _seg_vcl_intersection
    elif isinstance(objb, arc.Arc):
        _func = _arc_vcl_intersection
    elif isinstance(objb, (circle.Circle, ccircle.CCircle)):
        _func = _circ_vcl_intersection
    elif isinstance(objb, hcline.HCLine):
        _func = _hcl_vcl_intersection
    elif isinstance(objb, vcline.VCLine):
        _func = _non_intersecting
    elif isinstance(objb, acline.ACLine):
        _func = _vcl_acl_intersection
    elif isinstance(objb, cline.CLine):
        _func = _vcl_cl_intersection
    return _func

#
# set acline intersecting function
#

def _acline_intfuncs(objb):
    if isinstance(objb, segment.Segment):
        _func  = _seg_acl_intersection
    elif isinstance(objb, arc.Arc):
        _func = _arc_acl_intersection
    elif isinstance(objb, (circle.Circle, ccircle.CCircle)):
        _func = _circ_acl_intersection
    elif isinstance(objb, hcline.HCLine):
        _func = _hcl_acl_intersection
    elif isinstance(objb, vcline.VCLine):
        _func = _vcl_acl_intersection
    elif isinstance(objb, acline.ACLine):
        _func = _acl_acl_intersection
    elif isinstance(objb, cline.CLine):
        _func = _acl_cl_intersection
    else:
        _func = _null_intfunc
    return _func

#
# set cline intersecting function
#

def _cline_intfuncs(objb):
    if isinstance(objb, segment.Segment):
        _func  = _seg_cl_intersection
    elif isinstance(objb, arc.Arc):
        _func = _arc_cl_intersection
    elif isinstance(objb, (circle.Circle, ccircle.CCircle)):
        _func = _circ_cl_intersection
    elif isinstance(objb, hcline.HCLine):
        _func = _hcl_cl_intersection
    elif isinstance(objb, vcline.VCLine):
        _func = _vcl_cl_intersection
    elif isinstance(objb, acline.ACLine):
        _func = _acl_cl_intersection
    elif isinstance(objb, cline.CLine):
        _func = _cl_cl_intersection
    else:
        _func = _null_intfunc
    return _func

#
# intersection functions for polylines
#

def _get_intfunc(obja, objb):
    if isinstance(obja, segment.Segment):
        _intfunc = _segment_intfuncs(objb)
    elif isinstance(obja, arc.Arc):
        _intfunc = _arc_intfuncs(objb)
    elif isinstance(obja, (circle.Circle, ccircle.CCircle)):
        _intfunc = _circ_intfuncs(objb)
    elif isinstance(obja, hcline.HCLine):
        _intfunc = _hcline_intfuncs(objb)
    elif isinstance(obja, vcline.VCLine):
        _intfunc = _vcline_intfuncs(objb)
    elif isinstance(obja, acline.ACLine):
        _intfunc = _acline_intfuncs(objb)
    elif isinstance(obja, cline.CLine):
        _intfunc = _cline_intfuncs(objb)
    else:
        _intfunc = _null_intfunc
    return _intfunc

def find_intersections(obja, objb):
    _ipts = []
    _intfunc = _get_intfunc(obja, objb)
    _intfunc(_ipts, obja, objb)
    return _ipts



#
# Copyright (c) 2003, 2004 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# GTK interface code for dealing with text entities
#

import pygtk
pygtk.require('2.0')
import gtk
import pango

import Generic.text

def draw_textblock(gtkimage, tblock, active):
    _da = gtkimage.getDA()
    _gc = _da.get_style().fg_gc[gtk.STATE_NORMAL]
    _text = tblock.getText()
    _family = tblock.getFamily()
    _style = tblock.getStyle()
    _weight = tblock.getWeight()
    _color = tblock.getColor()
    _size = tblock.getSize()
    _layout = make_pango_layout(gtkimage, _text, _family, _style,
                                _weight, _color, _size)
    _x, _y = tblock.getLocation()
    _px, _py = gtkimage.coordToPixTransform(_x, _y)
    _pixmap = gtkimage.getPixmap()
    _pixmap.draw_layout(_gc, _px, _py, _layout)
    
def make_pango_layout(gtkimage, text, family, style, weight, color, size):
    _layout = pango.Layout(gtkimage.getDA().get_pango_context())
    _layout.set_text(text)
    _attlist = pango.AttrList()
    _family = pango.AttrFamily(family)
    _family.start_index = 0
    _family.end_index = -1
    _attlist.insert(_family)
    _sty = pango.AttrStyle(style)
    _sty.start_index = 0
    _sty.end_index = -1
    _attlist.insert(_sty)
    _wt = pango.AttrWeight(weight)
    _wt.start_index = 0
    _wt.end_index = -1
    _attlist.insert(_wt)
    _r, _g, _b = color.getColors()
    _rs = int(65535.0 * (_r/255.0))
    _gs = int(65535.0 * (_g/255.0))
    _bs = int(65535.0 * (_b/255.0))
    _fg = pango.AttrForeground(_rs, _gs, _bs)
    _fg.start_index = 0
    _fg.end_index = -1
    _attlist.insert(_fg)
    _sz = pango.AttrSize(1024 * size) # need to scale
    _sz.start_index = 0
    _sz.end_index = -1
    _attlist.insert(_sz)
    _layout.set_attributes(_attlist)
    return _layout
    
def add_text(gtkimage, tool):
    _text = tool.getText()
    _x, _y = tool.getTextLocation()
    _textstyle = gtkimage.getOption('TEXT_STYLE')
    _tblock = Generic.text.TextBlock(_text, _textstyle)
    _tblock.setLocation(_x, _y)
    gtkimage.addObject(_tblock)
    
def text_button_press(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _pt, _flag = gtkimage.findPoint(_x, _y, _tol)
    if _pt is not None:
        _x, _y = _pt.getCoords()
    tool.setTextLocation(_x, _y)
    tool.create(gtkimage)
    gtkimage.redraw()
    # tool.clearCurrentPoint()
    
def text_motion_notify(gtkimage, widget, event, tool):
    _tw, _th = tool.getPixelSize()
    _gc = gtkimage.getGC()
    _x = int(event.x)
    _y = int(event.y)
    _cp = tool.getCurrentPoint()
    if _cp is not None:
        _xc, _yc = _cp
        widget.window.draw_rectangle(_gc, gtk.FALSE, _xc, _yc, _tw, _th)
    tool.setCurrentPoint(_x, _y)
    widget.window.draw_rectangle(_gc, gtk.FALSE, _x, _y, _tw, _th)

def text_add_init(gtkimage, tool):
    _text = tool.getText()
    _family = gtkimage.getOption('FONT_FAMILY')
    _weight = gtkimage.getOption('FONT_WEIGHT')
    _style = gtkimage.getOption('FONT_STYLE')
    _size = gtkimage.getOption('FONT_SIZE')
    _color = gtkimage.getOption('FONT_COLOR')
    _layout = make_pango_layout(gtkimage, _text, _family, _style, _weight,
                                _color, _size)
    tool.setLayout(_layout)
    _lw, _lh = _layout.get_pixel_size()
    tool.setPixelSize(_lw, _lh)
    tool.setHandler("motion_notify", text_motion_notify)
    tool.setHandler("button_press", text_button_press)
    gtkimage.setPrompt("Click where to place the text")
    _gc = gtkimage.getGC()
    _gc.set_line_attributes(1, gtk.gdk.LINE_SOLID,
                            gtk.gdk.CAP_BUTT, gtk.gdk.JOIN_MITER)
    _gc.set_function(gtk.gdk.INVERT)
    
    
def make_text_dialog(gtkimage, text=None):
    _window = gtkimage.getWindow()
    _dialog = gtk.Dialog("Enter text", _window,
                         gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
                         (gtk.STOCK_OK, gtk.RESPONSE_OK,
                          gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL))
    _tb = gtk.TextBuffer()
    _tv = gtk.TextView(_tb)
    _sw = gtk.ScrolledWindow()
    _sw.set_size_request(400, 300)
    _sw.add_with_viewport(_tv)
    _dialog.vbox.pack_start(_sw, gtk.FALSE, gtk.FALSE, 0)
    _dialog.show_all()
    return _dialog, _tb

def get_new_text(gtkimage):
    _tool = gtkimage.getTool()
    _dialog, _tb = make_text_dialog(gtkimage)
    _response = _dialog.run()
    if _response == gtk.RESPONSE_OK:
        _start_iter, _end_iter = _tb.get_bounds()
        _text = _tb.get_text(_start_iter, _end_iter)
        if len(_text):
            _tool.setText(_text)
    _dialog.destroy()




#
# Copyright (c) 2002, 2003, 2004 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# classes for colors
#

import globals

#
# a seemingly good way to convert from string to integer
# for a couple of common ways of expressing colors ...
#

def color_str_to_int(cstr):
    if cstr.startswith('0x') or cstr.startswith('0X'):
        _val = int(cstr, 16)
    elif cstr.startswith('#'):
        _val = int(cstr[1:], 16)
    else:
        _val = int(cstr)
    return _val

class Color(object):
    """An object representing an RGB color.

The class purpose is self evident.

A Color object has three attributes:

r: Red (0-255)
g: Green (0-255)
b: Blue (0-255)

There is no alpha-channel attribute (yet)...

A Color object has the following methods:

getRed(): Get the Red value in the Color.
getBlue(): Get the Blue value in the Color.
getGreen(): Get the Green value in the Color.
clone(): Return an identical copy of a Color.

Once a color object is created, the values it
contains may not be changed.
    """
    def __init__(self, r=None, g=None, b=None):
        """Initialize a Color object.

There are several ways to create a color object:

Color(r,g,b) => r, g, and b values are all integers 0 <= value <= 255
Color(0xxxxx) => A hexidecimal value - it must begin with 0x. The color
                 is computed as follows:
                 r = (0xff0000 & value) >> 16
                 g = (0xff00 & value) >> 8
                 b = (0xff & value)
Color('#hexvalue') => A string prefixed with '#', with the remaining
                      characters representing a hexidecimal value.
Color() => A default color with r = g = b = 255.
        """
        _r = r
        _g = g
        _b = b
        if isinstance(_r, str):
            _val = color_str_to_int(_r)
            _r = (0xff0000 & _val) >> 16
            _g = (0xff00 & _val) >> 8
            _b = (0xff & _val)
        elif isinstance(_r, int) and _g is None and _b is None:
            _r = (0xff0000 & r) >> 16
            _g = (0xff00 & r) >> 8
            _b = (0xff & r)
        elif (isinstance(_r, int) and
              isinstance(_g, int) and
              isinstance(_b, int)):
            if _r < 0 or _r > 255:
                raise ValueError, "Invalid Red value: %d" % _r
            if _g < 0 or _g > 255:
                raise ValueError, "Invalid Green value: %d" % _g
            if _b < 0 or _b > 255:
                raise ValueError, "Invalid Blue value: %d" % _b
        elif _r is None and _g is None and _b is None:
            _r = 255
            _g = 255
            _b = 255
        else:
            raise SyntaxError, "Invalid call to Color()."
        self.__color = (_r << 16) | (_g << 8) | _b

    def __cmp__(self, obj):
        """Compare two Color objects.

The comparison is done based on the RGB values.

red value of C1 < red value of C2 ==> return -1
red value of C1 > red value of C2 ==> return 1

Then green values are compared, then blue. If
all values are equal, return 0.
        """
        if not isinstance(obj, Color):
            raise TypeError, "Invalid object for color comparison: " + `obj`
        _val = self.__color
        _objval = hash(obj)
        return cmp(_val, _objval)

    def __hash__(self):
        """Return a hash value for the Color object.

Providing this method means that Color objects can be used
as keys in dictionaries.
        """
        return self.__color
        
    def __repr__(self):
        _val = self.__color
        _r = (_val & 0xff0000) >> 16
        _g = (_val & 0xff00) >> 8
        _b = (_val & 0xff)
        return "Color(%d,%d,%d)" % (_r, _g, _b)
    
    def __str__(self):
        return "#%06x" % self.__color

    def getColors(self):
        """Return a three-item tuple with the values comprising this color.

getColors()        
        """
        _val = self.__color
        _r = (_val & 0xff0000) >> 16
        _g = (_val & 0xff00) >> 8
        _b = (_val & 0xff)
        return _r, _g, _b

    def getRed(self):
        """Return the red value of the color.

getRed()        
        """
        return (self.__color & 0xff0000) >> 16

    r = property(getRed, None, None, "Red value of the color.")

    def getGreen(self):
        """Return the green value of the color.

getGreen()        
        """
        return (self.__color & 0xff00) >> 8

    g = property(getGreen, None, None, "Green value of the color.")

    def getBlue(self):
        """Return the blue value of the color.

getBlue()        
        """
        return (self.__color & 0xff)

    b = property(getBlue, None, None, "Blue value of the color.")

    def clone(self):
        """Return a new Color object with the same color values.

clone()
        """
        _val = self.__color
        return Color(_val)

#
# ColorDict Class
#
# The ColorDict is built from the dict object. Using instances
# of this class will guarantee than only Color objects will be
# stored in the instance
#

class ColorDict(dict):
    def __init__(self):
        dict.__init__(self)
        
    def __setitem__(self, key, value):
        if not isinstance(key, Color):
            raise TypeError, "ColorDict keys must be color objects: " + `key`
        if not isinstance(value, Color):
            raise TypeError, "ColorDict values must be Color objects: " + `value`
        dict.__setitem__(self, key, value)

#
# find a Color object stored in the global color dictionary
# or make a new one and store it
#

def get_color(r, g, b):
    if not isinstance(r, int):
        raise TypeError, "Invalid red value:" + `r`
    if r < 0 or r > 255:
        raise ValueError, "Invalid red value: %d" % r
    if not isinstance(g, int):
        raise TypeError, "Invalid green value:" + `g`
    if g < 0 or g > 255:
        raise ValueError, "Invalid green value: %d" % g
    if not isinstance(b, int):
        raise TypeError, "Invalid blue value:" + `b`
    if b < 0 or b > 255:
        raise ValueError, "Invalid blue value: %d" % b
    _color = Color(r, g, b)
    if _color in globals.colors:
        _color = globals.colors[_color]
    else:
        globals.colors[_color] = _color
    return _color



#
# Copyright (c) 2002, 2003 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# code for setting the preferences of an image
#

import gtk

import Generic.text
import Generic.units

def set_dimension_prefs(gtkimage, prefdict):
    assert prefdict.has_key('OFFSET_ENTRY'), "Missing OFFSET_ENTRY"
    _text = prefdict['OFFSET_ENTRY'].get_chars(0, -1)
    if len(_text) and _text != '+':
        _offset = float(_text)
        gtkimage.setOption('DIM_OFFSET', _offset)
    assert prefdict.has_key('EXTENSION_ENTRY'), "Missing EXTENSION_ENTRY"
    _text = prefdict['EXTENSION_ENTRY'].get_chars(0, -1)
    if len(_text) and _text != '+':
        _extlen = float(_text)
        gtkimage.setOption('DIM_EXTENSION', _extlen)
    assert prefdict.has_key('DIM_POS_MENU'), "Missing DIM_POS_MENU"
    pos = prefdict['DIM_POS_MENU'].get_history()
    gtkimage.setOption('DIM_POSITION', pos)
    assert prefdict.has_key('DIM_EPT_MENU'), "Missing DIM_EPT_MENU"
    ept = prefdict['DIM_EPT_MENU'].get_history()
    gtkimage.setOption('DIM_ENDPOINT', ept)    
    assert prefdict.has_key('EPT_SIZE_ENTRY'), "Missing EPT_SIZE_ENTRY"
    _text = prefdict['EPT_SIZE_ENTRY'].get_chars(0, -1)
    if len(_text) and _text != '+':
        _size = float(_text)
        gtkimage.setOption('DIM_ENDPOINT_SIZE', _size)
    assert prefdict.has_key('DUAL_MODE_CB'), "Missing DUAL_MODE_CB"
    val = prefdict['DUAL_MODE_CB'].get_active()
    if val:
        state = True
    else:
        state = False
    gtkimage.setOption('DIM_DUAL_MODE', state)
    
    assert prefdict.has_key('PRIMARY_UNITS_MENU'), "Missing PRIMARY_UNITS_MENU"
    unit = prefdict['PRIMARY_UNITS_MENU'].get_history()
    gtkimage.setOption('DIM_PRIMARY_UNITS', unit)
    assert prefdict.has_key('PRIMARY_PREFIX_EB'), "Missing PRIMARY_PREFIX_EB"
    chars = prefdict['PRIMARY_PREFIX_EB'].get_chars(0,-1)
    if chars == "":
        pfx = u''
    else:
        try:
            pfx = unicode(chars)
        except:
            pfx = u'' # bailout?
    gtkimage.setOption('DIM_PRIMARY_PREFIX', pfx)
    assert prefdict.has_key('PRIMARY_SUFFIX_EB'), "Missing PRIMARY_SUFFIX_EB"
    chars = prefdict['PRIMARY_SUFFIX_EB'].get_chars(0,-1)
    if chars == "":
        sfx = u''
    else:
        try:
            sfx = unicode(chars)
        except:
            sfx = u'' # bailout?
    gtkimage.setOption('DIM_PRIMARY_SUFFIX', sfx)
    assert prefdict.has_key('PRIMARY_PREC_SB'), "Missing PRIMARY_PREC_SB"
    prec = prefdict['PRIMARY_PREC_SB'].get_value()
    gtkimage.setOption('DIM_PRIMARY_PRECISION', prec)
    assert prefdict.has_key('PRIMARY_LZ_CB'), 'Missing PRIMARY_LZ_CB'
    val = prefdict['PRIMARY_LZ_CB'].get_active()
    if val:
        state = True
    else:
        state = False
    gtkimage.setOption('DIM_PRIMARY_LEADING_ZERO', state)
    assert prefdict.has_key('PRIMARY_TD_CB'), 'Missing PRIMARY_TD_CB'
    val = prefdict['PRIMARY_TD_CB'].get_active()
    if val:
        state = True
    else:
        state = False
    gtkimage.setOption('DIM_PRIMARY_TRAILING_DECIMAL', state)

    assert prefdict.has_key('SECONDARY_UNITS_MENU'), "Missing SECONDARY_UNITS_MENU"
    unit = prefdict['SECONDARY_UNITS_MENU'].get_history()
    gtkimage.setOption('DIM_SECONDARY_UNITS', unit)
    assert prefdict.has_key('SECONDARY_PREFIX_EB'), "Missing SECONDARY_PREFIX_EB"
    chars = prefdict['SECONDARY_PREFIX_EB'].get_chars(0,-1)
    if chars == "":
        pfx = u''
    else:
        try:
            pfx = unicode(chars)
        except:
            pfx = u'' # bailout?
    gtkimage.setOption('DIM_SECONDARY_PREFIX', pfx)
    assert prefdict.has_key('SECONDARY_SUFFIX_EB'), "Missing SECONDARY_SUFFIX_EB"
    chars = prefdict['SECONDARY_SUFFIX_EB'].get_chars(0,-1)
    if chars == "":
        sfx = u''
    else:
        try:
            sfx = unicode(chars)
        except:
            sfx = u'' # bailout?
    gtkimage.setOption('DIM_SECONDARY_SUFFIX', sfx)
    assert prefdict.has_key('SECONDARY_PREC_SB'), "Missing SECONDARY_PREC_SB"
    prec = prefdict['SECONDARY_PREC_SB'].get_value()
    gtkimage.setOption('DIM_SECONDARY_PRECISION', prec)
    assert prefdict.has_key('SECONDARY_LZ_CB'), 'Missing SECONDARY_LZ_CB'
    val = prefdict['SECONDARY_LZ_CB'].get_active()
    if val:
        state = True
    else:
        state = False
    gtkimage.setOption('DIM_SECONDARY_LEADING_ZERO', state)
    assert prefdict.has_key('SECONDARY_TD_CB'), 'Missing SECONDARY_TD_CB'
    val = prefdict['SECONDARY_TD_CB'].get_active()
    if val:
        state = True
    else:
        state = False
    gtkimage.setOption('DIM_SECONDARY_TRAILING_DECIMAL', state)

    #
    # radial dimension prefix & suffix
    #
    
    assert prefdict.has_key('RADIAL_PRIMARY_PREFIX_EB'), "Missing RADIAL_PRIMARY_PREFIX_EB"
    chars = prefdict['RADIAL_PRIMARY_PREFIX_EB'].get_chars(0,-1)
    if chars == "":
        pfx = u''
    else:
        try:
            pfx = unicode(chars)
        except:
            pfx = u'' # bailout?
    gtkimage.setOption('RADIAL_DIM_PRIMARY_PREFIX', pfx)
    assert prefdict.has_key('RADIAL_SECONDARY_SUFFIX_EB'), "Missing RADIAL_SECONDARY_SUFFIX_EB"    
    chars = prefdict['RADIAL_PRIMARY_SUFFIX_EB'].get_chars(0,-1)
    if chars == "":
        sfx = u''
    else:
        try:
            sfx = unicode(chars)
        except:
            sfx = u'' # bailout?
    gtkimage.setOption('RADIAL_DIM_PRIMARY_SUFFIX', sfx)

    assert prefdict.has_key('RADIAL_SECONDARY_PREFIX_EB'), "Missing RADIAL_SECONDARY_PREFIX_EB"
    chars = prefdict['RADIAL_SECONDARY_PREFIX_EB'].get_chars(0,-1)
    if chars == "":
        pfx = u''
    else:
        try:
            pfx = unicode(chars)
        except:
            pfx = u'' # bailout?
    gtkimage.setOption('RADIAL_DIM_SECONDARY_PREFIX', pfx)
    assert prefdict.has_key('RADIAL_SECONDARY_SUFFIX_EB'), "Missing RADIAL_SECONDARY_SUFFIX_EB"    
    chars = prefdict['RADIAL_SECONDARY_SUFFIX_EB'].get_chars(0,-1)
    if chars == "":
        sfx = u''
    else:
        try:
            sfx = unicode(chars)
        except:
            sfx = u'' # bailout?
    gtkimage.setOption('RADIAL_DIM_SECONDARY_SUFFIX', sfx)

    #
    # angular dimension prefix & suffix
    #
    
    assert prefdict.has_key('ANGULAR_PRIMARY_PREFIX_EB'), "Missing ANGULAR_PRIMARY_PREFIX_EB"
    chars = prefdict['ANGULAR_PRIMARY_PREFIX_EB'].get_chars(0,-1)
    if chars == "":
        pfx = u''
    else:
        try:
            pfx = unicode(chars)
        except:
            pfx = u'' # bailout?
    gtkimage.setOption('ANGULAR_DIM_PRIMARY_PREFIX', pfx)
    assert prefdict.has_key('ANGULAR_SECONDARY_SUFFIX_EB'), "Missing ANGULAR_SECONDARY_SUFFIX_EB"    
    chars = prefdict['ANGULAR_PRIMARY_SUFFIX_EB'].get_chars(0,-1)
    if chars == "":
        sfx = u''
    else:
        try:
            sfx = unicode(chars)
        except:
            sfx = u'' # bailout?
    gtkimage.setOption('ANGULAR_DIM_PRIMARY_SUFFIX', sfx)

    assert prefdict.has_key('ANGULAR_SECONDARY_PREFIX_EB'), "Missing ANGULAR_SECONDARY_PREFIX_EB"
    chars = prefdict['ANGULAR_SECONDARY_PREFIX_EB'].get_chars(0,-1)
    if chars == "":
        pfx = u''
    else:
        try:
            pfx = unicode(chars)
        except:
            pfx = u'' # bailout?
    gtkimage.setOption('ANGULAR_DIM_SECONDARY_PREFIX', pfx)
    assert prefdict.has_key('ANGULAR_SECONDARY_SUFFIX_EB'), "Missing ANGULAR_SECONDARY_SUFFIX_EB"    
    chars = prefdict['ANGULAR_SECONDARY_SUFFIX_EB'].get_chars(0,-1)
    if chars == "":
        sfx = u''
    else:
        try:
            sfx = unicode(chars)
        except:
            sfx = u'' # bailout?
    gtkimage.setOption('ANGULAR_DIM_SECONDARY_SUFFIX', sfx)

    #
    # fonts
    #

    assert prefdict.has_key('PRIMARY_FONT_SEL'), "Missing PRIMARY_FONT_SEL"
    fn = prefdict['PRIMARY_FONT_SEL'].get_font_name()
    family, style, weight, stretch, size = Generic.text.parse_font(fn)
    gtkimage.setOption('DIM_PRIMARY_FONT_FAMILY', family)
    gtkimage.setOption('DIM_PRIMARY_FONT_SIZE', size)
    gtkimage.setOption('DIM_PRIMARY_FONT_STYLE', style)
    gtkimage.setOption('DIM_PRIMARY_FONT_WEIGHT', weight)

    assert prefdict.has_key('SECONDARY_FONT_SEL'), "Missing SECONDARY_FONT_SEL"
    fn = prefdict['SECONDARY_FONT_SEL'].get_font_name()
    family, style, weight, stretch, size = Generic.text.parse_font(fn)
    gtkimage.setOption('DIM_SECONDARY_FONT_FAMILY', family)
    gtkimage.setOption('DIM_SECONDARY_FONT_SIZE', size)
    gtkimage.setOption('DIM_SECONDARY_FONT_STYLE', style)
    gtkimage.setOption('DIM_SECONDARY_FONT_WEIGHT', weight)

#
# common routines for entry box validation
#
# eventually the preferences will be combined into a single
# dialog display so the duplication of code here and in gtkprefs.py
# will go away ...
#

def _entry_move_cursor(entry):
    entry.set_position(-1)
    return gtk.FALSE

def _entry_activate(entry): # this probably isn't a good choice ...
    entry.stop_emission("activate")

def _entry_focus_out(entry, event):
    _text = entry.get_chars(0, -1)
    if _text == '-' or _text == '+':
        entry.delete_text(0, -1)
    return gtk.FALSE

def _entry_insert_text(entry, new_text, new_text_length, position):
    if (new_text.isdigit() or
        new_text == '.' or
        new_text == '+'):
        _string = entry.get_chars(0, -1) + new_text
        _hid = entry.get_data('handlerid')
        entry.handler_block(_hid)
        _pos = entry.get_position()
        _move = True
        if _string == '+':
            _pos = entry.insert_text(new_text, _pos)
        else:
            try:
                _val = float(_string)
                _pos = entry.insert_text(new_text, _pos)
            except StandardError, e:
                _move = False
        entry.handler_unblock(_hid)
        gtk.idle_add(_entry_move_cursor, entry)
    entry.stop_emission("insert-text")

#

def add_secondary_font_page(gtkimage, pd, nb):
    vbox = gtk.VBox(gtk.FALSE, 2)
    frame = gtk.Frame("Secondary Dimension Font Properties")
    vbox.pack_start(frame, gtk.FALSE, gtk.FALSE)
    
    fw = gtk.FontSelection()
    fontlist = []
    fontlist.append(gtkimage.getOption('DIM_SECONDARY_FONT_FAMILY'))
    weight = gtkimage.getOption('DIM_SECONDARY_FONT_WEIGHT')
    if weight != Generic.text.NORMAL:
        if weight == 1:
            fontlist.append('Light')
        elif weight == 2:
            fontlist.append('Bold')
        elif weight == 3:
            fontlist.append('Heavy')
    style = gtkimage.getOption('DIM_SECONDARY_FONT_STYLE')
    if style != Generic.text.NORMAL:
        if style == 1:
            fontlist.append('Oblique')
        elif style == 2:
            fontlist.append('Italic')
    fontlist.append(str(gtkimage.getOption('DIM_SECONDARY_FONT_SIZE')))
    fw.set_font_name(' '.join(fontlist))
    pd['SECONDARY_FONT_SEL'] = fw
    frame.add(fw)
    vbox.show_all()

    lb = gtk.Label("Secondary Fonts")
    nb.append_page(vbox, lb)

def add_primary_font_page(gtkimage, pd, nb):
    vbox = gtk.VBox(gtk.FALSE, 2)
    frame = gtk.Frame("Primary Dimension Font Properties")
    vbox.pack_start(frame, gtk.FALSE, gtk.FALSE)
    
    fw = gtk.FontSelection()
    fontlist = []
    fontlist.append(gtkimage.getOption('DIM_PRIMARY_FONT_FAMILY'))
    weight = gtkimage.getOption('DIM_PRIMARY_FONT_WEIGHT')
    if weight != Generic.text.NORMAL:
        if weight == 1:
            fontlist.append('Light')
        elif weight == 2:
            fontlist.append('Bold')
        elif weight == 3:
            fontlist.append('Heavy')
    style = gtkimage.getOption('DIM_PRIMARY_FONT_STYLE')
    if style != Generic.text.NORMAL:
        if style == 1:
            fontlist.append('Oblique')
        elif style == 2:
            fontlist.append('Italic')
    fontlist.append(str(gtkimage.getOption('DIM_PRIMARY_FONT_SIZE')))
    fw.set_font_name(' '.join(fontlist))
    pd['PRIMARY_FONT_SEL'] = fw
    frame.add(fw)
    vbox.show_all()

    lb = gtk.Label("Primary Fonts")
    nb.append_page(vbox, lb)

def add_secondary_page(gtkimage, pd, nb):
    vbox = gtk.VBox(gtk.FALSE, 2)
    frame = gtk.Frame("Secondary Dimension")
    vbox.pack_start(frame, gtk.FALSE, gtk.FALSE)

    fvbox = gtk.VBox(gtk.FALSE, 2)
    frame.add(fvbox)
    
    hbox = gtk.HBox(gtk.FALSE, 2)
    ml = gtk.Label("Units:")
    hbox.pack_start(ml, gtk.FALSE, gtk.FALSE)
    menu = gtk.Menu()
    idx = 0
    unit_list = Generic.units.get_all_units()
    cur_unit = gtkimage.getOption('DIM_SECONDARY_UNITS')
    for i in xrange(len(unit_list)):
        if i == cur_unit:
            idx = i
        item = gtk.MenuItem(unit_list[i])
        menu.append(item)
    opt_menu = gtk.OptionMenu()
    opt_menu.set_menu(menu)
    opt_menu.set_history(idx)
    pd['SECONDARY_UNITS_MENU'] = opt_menu
    hbox.pack_start(opt_menu, gtk.FALSE, gtk.FALSE)
    fvbox.pack_start(hbox, gtk.FALSE, gtk.FALSE)

    hbox = gtk.HBox(gtk.FALSE, 2)
    lb = gtk.Label("Default Prefix:")
    hbox.pack_start(lb, gtk.FALSE, gtk.FALSE)

    eb = gtk.Entry()
    prefix = gtkimage.getOption('DIM_SECONDARY_PREFIX')
    eb.set_text(prefix)
    pd['SECONDARY_PREFIX_EB'] = eb
    hbox.pack_start(eb, gtk.TRUE, gtk.TRUE, 0)
    fvbox.pack_start(hbox, gtk.FALSE, gtk.FALSE)
    
    hbox = gtk.HBox(gtk.FALSE, 2)
    lb = gtk.Label("Default Suffix:")
    hbox.pack_start(lb, gtk.FALSE, gtk.FALSE)
    eb = gtk.Entry()
    suffix = gtkimage.getOption('DIM_SECONDARY_SUFFIX')
    eb.set_text(suffix)
    pd['SECONDARY_SUFFIX_EB'] = eb
    hbox.pack_start(eb, gtk.TRUE, gtk.TRUE, 0)
    fvbox.pack_start(hbox, gtk.FALSE, gtk.FALSE)

    hbox = gtk.HBox(gtk.FALSE, 2)
    lb = gtk.Label("Display precision:")
    hbox.pack_start(lb, gtk.FALSE, gtk.FALSE)
    prec = gtkimage.getOption('DIM_SECONDARY_PRECISION')
    adj = gtk.Adjustment(prec, 0, 15, 1, 1, 1)
    sb = gtk.SpinButton(adj)
    sb.set_digits(0)
    sb.set_numeric(gtk.TRUE)
    pd['SECONDARY_PREC_SB'] = sb
    hbox.pack_start(sb, gtk.TRUE, gtk.TRUE, 0)
    fvbox.pack_start(hbox, gtk.FALSE, gtk.FALSE)

    cb = gtk.CheckButton("Print leading 0")
    state = gtkimage.getOption('DIM_SECONDARY_LEADING_ZERO')
    cb.set_active(state)
    pd['SECONDARY_LZ_CB'] = cb
    fvbox.pack_start(cb, gtk.FALSE, gtk.FALSE)

    cb = gtk.CheckButton("Print trailing decimal point")
    state = gtkimage.getOption('DIM_SECONDARY_TRAILING_DECIMAL')
    cb.set_active(state)
    pd['SECONDARY_TD_CB'] = cb
    fvbox.pack_start(cb, gtk.FALSE, gtk.FALSE)

    vbox.show_all()
    
    tl = gtk.Label("Secondary")
    nb.append_page(vbox, tl)

def add_angular_page(gtkimage, pd, nb):
    vbox = gtk.VBox(gtk.FALSE, 2)
    frame = gtk.Frame("Angular Dimension Prefix & Suffix")
    vbox.pack_start(frame, gtk.FALSE, gtk.FALSE)

    fvbox = gtk.VBox(gtk.FALSE, 2)
    frame.add(fvbox)

    hbox = gtk.HBox(gtk.FALSE, 2)
    lb = gtk.Label("Default Primary Prefix:")
    hbox.pack_start(lb, gtk.FALSE, gtk.FALSE)

    eb = gtk.Entry()
    prefix = gtkimage.getOption('ANGULAR_DIM_PRIMARY_PREFIX')
    eb.set_text(prefix)
    pd['ANGULAR_PRIMARY_PREFIX_EB'] = eb
    hbox.pack_start(eb, gtk.TRUE, gtk.TRUE, 0)
    fvbox.pack_start(hbox, gtk.FALSE, gtk.FALSE)
    
    hbox = gtk.HBox(gtk.FALSE, 2)
    lb = gtk.Label("Default Primary Suffix:")
    hbox.pack_start(lb, gtk.FALSE, gtk.FALSE)
    
    eb = gtk.Entry()
    suffix = gtkimage.getOption('ANGULAR_DIM_PRIMARY_SUFFIX')
    eb.set_text(suffix)
    pd['ANGULAR_PRIMARY_SUFFIX_EB'] = eb
    hbox.pack_start(eb, gtk.TRUE, gtk.TRUE, 0)
    fvbox.pack_start(hbox, gtk.FALSE, gtk.FALSE)

    hbox = gtk.HBox(gtk.FALSE, 2)
    lb = gtk.Label("Default Secondary Prefix:")
    hbox.pack_start(lb, gtk.FALSE, gtk.FALSE)

    eb = gtk.Entry()
    prefix = gtkimage.getOption('ANGULAR_DIM_SECONDARY_PREFIX')
    eb.set_text(prefix)
    pd['ANGULAR_SECONDARY_PREFIX_EB'] = eb
    hbox.pack_start(eb, gtk.TRUE, gtk.TRUE, 0)
    fvbox.pack_start(hbox, gtk.FALSE, gtk.FALSE)
    
    hbox = gtk.HBox(gtk.FALSE, 2)
    lb = gtk.Label("Default Secondary Suffix:")
    hbox.pack_start(lb, gtk.FALSE, gtk.FALSE)
    
    eb = gtk.Entry()
    suffix = gtkimage.getOption('ANGULAR_DIM_SECONDARY_SUFFIX')
    eb.set_text(suffix)
    pd['ANGULAR_SECONDARY_SUFFIX_EB'] = eb
    hbox.pack_start(eb, gtk.TRUE, gtk.TRUE, 0)
    fvbox.pack_start(hbox, gtk.FALSE, gtk.FALSE)

    tl = gtk.Label("Angular")
    nb.append_page(vbox, tl)

def add_radial_page(gtkimage, pd, nb):
    vbox = gtk.VBox(gtk.FALSE, 2)
    frame = gtk.Frame("Radial Dimension Prefix & Suffix")
    vbox.pack_start(frame, gtk.FALSE, gtk.FALSE)

    fvbox = gtk.VBox(gtk.FALSE, 2)
    frame.add(fvbox)

    hbox = gtk.HBox(gtk.FALSE, 2)
    lb = gtk.Label("Default Primary Prefix:")
    hbox.pack_start(lb, gtk.FALSE, gtk.FALSE)

    eb = gtk.Entry()
    prefix = gtkimage.getOption('RADIAL_DIM_PRIMARY_PREFIX')
    eb.set_text(prefix)
    pd['RADIAL_PRIMARY_PREFIX_EB'] = eb
    hbox.pack_start(eb, gtk.TRUE, gtk.TRUE, 0)
    fvbox.pack_start(hbox, gtk.FALSE, gtk.FALSE)
    
    hbox = gtk.HBox(gtk.FALSE, 2)
    lb = gtk.Label("Default Primary Suffix:")
    hbox.pack_start(lb, gtk.FALSE, gtk.FALSE)
    
    eb = gtk.Entry()
    suffix = gtkimage.getOption('RADIAL_DIM_PRIMARY_SUFFIX')
    eb.set_text(suffix)
    pd['RADIAL_PRIMARY_SUFFIX_EB'] = eb
    hbox.pack_start(eb, gtk.TRUE, gtk.TRUE, 0)
    fvbox.pack_start(hbox, gtk.FALSE, gtk.FALSE)

    hbox = gtk.HBox(gtk.FALSE, 2)
    lb = gtk.Label("Default Secondary Prefix:")
    hbox.pack_start(lb, gtk.FALSE, gtk.FALSE)

    eb = gtk.Entry()
    prefix = gtkimage.getOption('RADIAL_DIM_SECONDARY_PREFIX')
    eb.set_text(prefix)
    pd['RADIAL_SECONDARY_PREFIX_EB'] = eb
    hbox.pack_start(eb, gtk.TRUE, gtk.TRUE, 0)
    fvbox.pack_start(hbox, gtk.FALSE, gtk.FALSE)
    
    hbox = gtk.HBox(gtk.FALSE, 2)
    lb = gtk.Label("Default Secondary Suffix:")
    hbox.pack_start(lb, gtk.FALSE, gtk.FALSE)
    
    eb = gtk.Entry()
    suffix = gtkimage.getOption('RADIAL_DIM_SECONDARY_SUFFIX')
    eb.set_text(suffix)
    pd['RADIAL_SECONDARY_SUFFIX_EB'] = eb
    hbox.pack_start(eb, gtk.TRUE, gtk.TRUE, 0)
    fvbox.pack_start(hbox, gtk.FALSE, gtk.FALSE)

    tl = gtk.Label("Radial")
    nb.append_page(vbox, tl)
    
def add_primary_page(gtkimage, pd, nb):
    vbox = gtk.VBox(gtk.FALSE, 2)
    frame = gtk.Frame("Primary Dimension")
    vbox.pack_start(frame, gtk.FALSE, gtk.FALSE)

    fvbox = gtk.VBox(gtk.FALSE, 2)
    frame.add(fvbox)
    
    hbox = gtk.HBox(gtk.FALSE, 2)
    ml = gtk.Label("Units:")
    hbox.pack_start(ml, gtk.FALSE, gtk.FALSE)
    menu = gtk.Menu()
    idx = 0
    unit_list = Generic.units.get_all_units()
    cur_unit = gtkimage.getOption('DIM_PRIMARY_UNITS')
    for i in xrange(len(unit_list)):
        if i == cur_unit:
            idx = i
        item = gtk.MenuItem(unit_list[i])
        menu.append(item)
    opt_menu = gtk.OptionMenu()
    opt_menu.set_menu(menu)
    opt_menu.set_history(idx)
    pd['PRIMARY_UNITS_MENU'] = opt_menu
    hbox.pack_start(opt_menu, gtk.FALSE, gtk.FALSE)
    fvbox.pack_start(hbox, gtk.FALSE, gtk.FALSE)

    hbox = gtk.HBox(gtk.FALSE, 2)
    lb = gtk.Label("Default Prefix:")
    hbox.pack_start(lb, gtk.FALSE, gtk.FALSE)

    eb = gtk.Entry()
    prefix = gtkimage.getOption('DIM_PRIMARY_PREFIX')
    eb.set_text(prefix)
    pd['PRIMARY_PREFIX_EB'] = eb
    hbox.pack_start(eb, gtk.TRUE, gtk.TRUE, 0)
    fvbox.pack_start(hbox, gtk.FALSE, gtk.FALSE)
    
    hbox = gtk.HBox(gtk.FALSE, 2)
    lb = gtk.Label("Default Suffix:")
    hbox.pack_start(lb, gtk.FALSE, gtk.FALSE)
    
    eb = gtk.Entry()
    suffix = gtkimage.getOption('DIM_PRIMARY_SUFFIX')
    eb.set_text(suffix)
    pd['PRIMARY_SUFFIX_EB'] = eb
    hbox.pack_start(eb, gtk.TRUE, gtk.TRUE, 0)
    fvbox.pack_start(hbox, gtk.FALSE, gtk.FALSE)

    hbox = gtk.HBox(gtk.FALSE, 2)
    lb = gtk.Label("Display precision:")
    hbox.pack_start(lb, gtk.FALSE, gtk.FALSE)
    prec = gtkimage.getOption('DIM_PRIMARY_PRECISION')
    adj = gtk.Adjustment(prec, 0, 15, 1, 1, 1)
    sb = gtk.SpinButton(adj)
    sb.set_digits(0)
    sb.set_numeric(gtk.TRUE)
    pd['PRIMARY_PREC_SB'] = sb
    hbox.pack_start(sb, gtk.TRUE, gtk.TRUE, 0)
    fvbox.pack_start(hbox, gtk.FALSE, gtk.FALSE)

    cb = gtk.CheckButton("Print leading 0")
    state = gtkimage.getOption('DIM_PRIMARY_LEADING_ZERO')
    cb.set_active(state)
    pd['PRIMARY_LZ_CB'] = cb
    fvbox.pack_start(cb, gtk.FALSE, gtk.FALSE)

    cb = gtk.CheckButton("Print trailing decimal point")
    state = gtkimage.getOption('DIM_PRIMARY_TRAILING_DECIMAL')
    cb.set_active(state)
    pd['PRIMARY_TD_CB'] = cb
    fvbox.pack_start(cb, gtk.FALSE, gtk.FALSE)

    vbox.show_all()
    
    tl = gtk.Label("Primary")
    nb.append_page(vbox, tl)

def add_common_page(gtkimage, pd, nb):
    _vbox = gtk.VBox(gtk.FALSE, 2)
    _frame = gtk.Frame("General Properties")
    _vbox.pack_start(_frame, gtk.FALSE, gtk.FALSE)

    _fvbox = gtk.VBox(gtk.FALSE, 2)
    _frame.add(_fvbox)

    _hbox = gtk.HBox(gtk.FALSE, 2)
    _label = gtk.Label("Offset length:")
    _hbox.pack_start(_label, gtk.FALSE, gtk.FALSE, 0)
    _offset = "%g" % gtkimage.getOption('DIM_OFFSET')
    _entry = gtk.Entry()
    _entry.set_text(_offset)
    _entry.connect("focus_out_event", _entry_focus_out)
    _entry.connect("activate", _entry_activate)
    _handlerid = _entry.connect("insert-text", _entry_insert_text)
    _entry.set_data('handlerid', _handlerid)
    pd['OFFSET_ENTRY'] = _entry
    _hbox.pack_start(_entry, gtk.TRUE, gtk.TRUE, 0)
    _fvbox.pack_start(_hbox, gtk.FALSE, gtk.FALSE, 0)

    _hbox = gtk.HBox(gtk.FALSE, 2)
    _label = gtk.Label("Extension length:")
    _hbox.pack_start(_label, gtk.FALSE, gtk.FALSE, 0)
    _extlen = "%g" % gtkimage.getOption('DIM_EXTENSION')
    _entry = gtk.Entry()
    _entry.set_text(_extlen)
    _entry.connect("focus_out_event", _entry_focus_out)
    _entry.connect("activate", _entry_activate)
    _handlerid = _entry.connect("insert-text", _entry_insert_text)
    _entry.set_data('handlerid', _handlerid)
    pd['EXTENSION_ENTRY'] = _entry
    _hbox.pack_start(_entry, gtk.TRUE, gtk.TRUE, 0)
    _fvbox.pack_start(_hbox, gtk.FALSE, gtk.FALSE, 0)

    _hbox = gtk.HBox(gtk.FALSE, 2)
    _label = gtk.Label("Dimension Position:")
    _hbox.pack_start(_label, gtk.FALSE, gtk.FALSE, 0)
    _menu = gtk.Menu()
    _idx = 0
    _opts = ['Split Line', 'Above Line', 'Below Line']
    _cur_pos = gtkimage.getOption('DIM_POSITION')
    for _i in xrange(len(_opts)):
        if _i == _cur_pos:
            _idx = _i
        _item = gtk.MenuItem(_opts[_i])
        _menu.append(_item)
    _opt_menu = gtk.OptionMenu()
    _opt_menu.set_menu(_menu)
    _opt_menu.set_history(_idx)
    pd['DIM_POS_MENU'] = _opt_menu
    _hbox.pack_start(_opt_menu, gtk.FALSE, gtk.FALSE, 0)
    _fvbox.pack_start(_hbox, gtk.FALSE, gtk.FALSE, 0)

    _hbox = gtk.HBox(gtk.FALSE, 2)
    _label = gtk.Label("Dimension Endpoint Style:")
    _hbox.pack_start(_label, gtk.FALSE, gtk.FALSE, 0)
    _menu = gtk.Menu()
    _idx = 0
    _opts = ['None', 'Arrow', 'Filled Arrow', 'Slash', 'Circle']
    _cur_ept = gtkimage.getOption('DIM_ENDPOINT')
    for _i in xrange(len(_opts)):
        if _i == _cur_ept:
            _idx = _i
        _item = gtk.MenuItem(_opts[_i])
        _menu.append(_item)
    _opt_menu = gtk.OptionMenu()
    _opt_menu.set_menu(_menu)
    _opt_menu.set_history(_idx)
    pd['DIM_EPT_MENU'] = _opt_menu
    _hbox.pack_start(_opt_menu, gtk.FALSE, gtk.FALSE, 0)
    _fvbox.pack_start(_hbox, gtk.FALSE, gtk.FALSE, 0)

    _hbox = gtk.HBox(gtk.FALSE, 2)
    _lb = gtk.Label("Endpoint Size:")
    _hbox.pack_start(_lb, gtk.FALSE, gtk.FALSE)
    _size = "%g" % gtkimage.getOption('DIM_ENDPOINT_SIZE')
    _entry = gtk.Entry()
    _entry.set_text(_size)
    _entry.connect("focus_out_event", _entry_focus_out)
    _entry.connect("activate", _entry_activate)
    _handlerid = _entry.connect("insert-text", _entry_insert_text)
    _entry.set_data('handlerid', _handlerid)
    pd['EPT_SIZE_ENTRY'] = _entry
    _hbox.pack_start(_entry, gtk.TRUE, gtk.TRUE, 0)
    _fvbox.pack_start(_hbox, gtk.FALSE, gtk.FALSE, 0)

    _cb = gtk.CheckButton("Display Secondary Dimension")
    _state = gtkimage.getOption('DIM_DUAL_MODE')
    _cb.set_active(_state)
    pd['DUAL_MODE_CB'] = _cb
    _fvbox.pack_start(_cb, gtk.FALSE, gtk.FALSE, 0)

    _label = gtk.Label("General")
    nb.append_page(_vbox, _label)

def check_dim_prefs(gtkimage):
    window = gtkimage.getWindow()
    dialog = gtk.Dialog("Dimension Preference", window,
                        gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
                        (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                         gtk.STOCK_APPLY, gtk.RESPONSE_APPLY,
                         gtk.STOCK_OK, gtk.RESPONSE_OK))
    notebook = gtk.Notebook()
    pref_dict = {}
    dialog.vbox.pack_start(notebook, expand=gtk.TRUE)
    add_common_page(gtkimage, pref_dict, notebook)
    add_primary_page(gtkimage, pref_dict, notebook)
    add_secondary_page(gtkimage, pref_dict, notebook)
    add_radial_page(gtkimage, pref_dict, notebook)
    add_angular_page(gtkimage, pref_dict, notebook)
    add_primary_font_page(gtkimage, pref_dict, notebook)
    add_secondary_font_page(gtkimage, pref_dict, notebook)
    notebook.set_current_page(0)
    dialog.show_all()
    response = dialog.run()
    if response == gtk.RESPONSE_OK or response == gtk.RESPONSE_APPLY:
        print "got RESPONSE_OK or RESPONSE_APPLY"
        set_dimension_prefs(gtkimage, pref_dict)
    else:
        print "got RESPONSE_CANCEL"
    dialog.destroy()



#
# Copyright (c) 2002, 2003, 2004 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# functions for doing mirroring operations
#

import pygtk
pygtk.require('2.0')
import gtk

import Generic.hcline
import Generic.acline
import Generic.vcline
import Generic.cline

import Generic.mirror as mirror

def mirror_objects(gtkimage, tool, objs):
    _mline = tool.getMirrorLine()
    if len(objs):
        gtkimage.startAction()
        try:
            _mobjs = mirror.mirror_objects(gtkimage, _mline, objs)
        finally:
            gtkimage.endAction()
    gtkimage.redraw()
    gtkimage.setPrompt("Click on the mirroring construction line.")
    tool.reset()
    mirror_mode_init(tool)
    
def select_motion_notify(gtkimage, widget, event, tool):
    _tx, _ty = tool.getLocation()
    _px, _py = gtkimage.coordToPixTransform(_tx, _ty)
    _gc = gtkimage.getGC()
    _x = int(event.x)
    _y = int(event.y)
    _cp = tool.getCurrentPoint()
    if _cp is not None:
        _xc, _yc = _cp
        _xmin = min(_xc, _px)
        _ymin = min(_yc, _py)
        _rw = abs(_xc - _px)
        _rh = abs(_yc - _py)
        widget.window.draw_rectangle(_gc, gtk.FALSE, _xmin, _ymin, _rw, _rh)
    tool.setCurrentPoint(_x, _y)
    _xmin = min(_x, _px)
    _ymin = min(_y, _py)
    _rw = abs(_x - _px)
    _rh = abs(_y - _py)
    widget.window.draw_rectangle(_gc, gtk.FALSE, _xmin, _ymin, _rw, _rh)

def second_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _active_layer = gtkimage.getActiveLayer()
    _p = _active_layer.find('point', _x, _y)
    if _p is not None:
        _x, _y = _p.getCoords()
    _x1, _y1 = tool.getLocation()
    _xmin = min(_x1, _x)
    _ymin = min(_y1, _y)
    _xmax = max(_x1, _x)
    _ymax = max(_y1, _y)
    _objs = _active_layer.objsInRegion(_xmin, _ymin, _xmax, _ymax, True)
    mirror_objects(gtkimage, tool, _objs)

def first_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _objdict = gtkimage.mapPoint(_x, _y, _tol)
    if len(_objdict):
        _active_layer = gtkimage.getActiveLayer()
        if _active_layer in _objdict:
            _objs = []
            for _obj, _pt in _objdict[_active_layer]:
                _objs.append(_obj)
            _mline = tool.getMirrorLine()
            mirror_objects(gtkimage, tool, _objs)
    else:
        _p, _flag = gtkimage.findPoint(_x, _y, _tol)
        if _p is not None:
            _x, _y = _p.getCoords()
        tool.setLocation(_x, _y)
        tool.setHandler("motion_notify", select_motion_notify)
        tool.setHandler("button_press", second_button_press_cb)
        gtkimage.getGC().set_function(gtk.gdk.INVERT)        
        
def get_mirror_line_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _objdict = gtkimage.mapPoint(_x, _y, _tol)
    if len(_objdict):
        _active_layer = gtkimage.getActiveLayer()
        _got_line = False
        if _active_layer in _objdict:
            for _obj, _pt in _objdict[_active_layer]:
                if isinstance(_obj, (Generic.hcline.HCLine,
                                     Generic.vcline.VCLine,
                                     Generic.acline.ACLine,
                                     Generic.cline.CLine)):
                    tool.setHandler("button_press", first_button_press_cb)
                    tool.setMirrorLine(_obj)
                    gtkimage.setPrompt("Click on the objects to mirror.")
                    break

def mirror_mode_init(tool):
    tool.setHandler("button_press", get_mirror_line_cb)



#
# Copyright (c) 2002-2004 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# main routine to start Cocoa-based PythonCad
#

import getopt
import sys
import os

import Generic.globals
import Generic.image
import Generic.imageio
import Generic.fileio
import Generic.selections
import Generic.preferences
import Generic.color
import Generic.linetype
import Generic.style
import Generic.segment
import Generic.circle
import Generic.leader
import Generic.polyline
import Generic.segjoint
import Generic.conobject
import Generic.baseobject
import Generic.dimension
import Generic.units
import Interface.Cocoa.ImageDocument 
import Interface.Cocoa.CADView
import Interface.Cocoa.AppController

from PyObjCTools import NibClassBuilder, AppHelper

#Initialize NIB
NibClassBuilder.extractClasses("ImageDocument")

#
# set up global variables ...
#

Generic.globals.prefs = Generic.baseobject.LockedDict(str)
Generic.globals.colors = Generic.baseobject.TypedDict(Generic.color.Color,
                                                      Generic.color.Color)
Generic.globals.linetypes = Generic.baseobject.TypedDict(Generic.linetype.Linetype, Generic.linetype.Linetype)
Generic.globals.styles = Generic.baseobject.TypedDict(Generic.style.Style,
                                                      Generic.style.Style)
Generic.globals.dimstyles = []

Generic.globals.selectobj = Generic.selections.Selection()


def _initialize_booleans():
    Generic.globals.prefs['HIGHLIGHT_POINTS'] = True
    Generic.globals.prefs['AUTOSPLIT'] = False

def _initialize_sizes():
    Generic.globals.prefs['CHAMFER_LENGTH'] = 1.0
    Generic.globals.prefs['FILLET_RADIUS'] = 1.0
    Generic.globals.prefs['UNITS'] = Generic.units.MILLIMETERS
    Generic.globals.prefs['LEADER_ARROW_SIZE'] = 1.0

def _initialize_image_colors():
    _color = Generic.color.get_color(80, 140, 210) # blueish/purpleish
    Generic.globals.prefs['INACTIVE_LAYER_COLOR'] = _color
    _color = Generic.color.get_color(0, 0, 0) # black
    Generic.globals.prefs['BACKGROUND_COLOR'] = _color
    _color = Generic.color.get_color(255, 255, 0) #yellow
    Generic.globals.prefs['SINGLE_POINT_COLOR'] = _color
    _color = Generic.color.get_color(0, 255, 255) #cyan
    Generic.globals.prefs['MULTI_POINT_COLOR'] = _color

def _initialize_styles():
    _color = Generic.color.get_color(0xff, 0xff, 0xff) # white
    _lt = Generic.linetype.get_linetype_by_dashes(None)
    _style = Generic.style.Style(u'Solid White Line', _lt, _color, 1.0)
    Generic.globals.styles[_style] = _style
    Generic.globals.prefs['LINE_STYLE'] = _style
    Generic.globals.prefs['LINE_COLOR'] = _style.getColor()
    Generic.globals.prefs['LINE_TYPE'] = _style.getLinetype()
    Generic.globals.prefs['LINE_THICKNESS'] = _style.getThickness()
    #
    # set this style as the class default for the "real" drawing entities
    #
    # fixme: this should be done with a classmethod or some sort of
    # function ...
    #
    Generic.segment.Segment.__defstyle = _style
    Generic.circle.Circle.__defstyle = _style
    Generic.leader.Leader.__defstyle = _style
    Generic.polyline.Polyline.__defstyle = _style
    Generic.segjoint.SegJoint.__defstyle = _style
    #
    _color = Generic.color.get_color(0, 0, 0) # black
    _style = Generic.style.Style(u'Solid Black Line', _lt, _color, 1.0)
    Generic.globals.styles[_style] = _style
    _color = Generic.color.get_color(0xff, 0, 0) # red
    _lt = Generic.linetype.get_linetype_by_dashes([4, 1])
    _style = Generic.style.Style(u'Dashed Red Line', _lt, _color, 1.0)
    Generic.globals.styles[_style] = _style
    _color = Generic.color.get_color(0, 0xff, 0) # green
    _style = Generic.style.Style(u'Dashed Green Line', _lt, _color, 1.0)
    Generic.globals.styles[_style] = _style
    _color = Generic.color.get_color(0, 0, 0xff) # blue
    _style = Generic.style.Style(u'Dashed Blue Line', _lt, _color, 1.0)
    Generic.globals.styles[_style] = _style
    _color = Generic.color.get_color(0xff, 0xff, 0) # yellow
    _lt = Generic.linetype.get_linetype_by_dashes([8, 2])
    _style = Generic.style.Style(u'Dashed Yellow Line', _lt, _color, 1.0)
    Generic.globals.styles[_style] = _style
    _color = Generic.color.get_color(0xff, 0, 0xff) # violet
    _style = Generic.style.Style(u'Dashed Violet Line', _lt, _color, 1.0)
    Generic.globals.styles[_style] = _style
    _color = Generic.color.get_color(0, 0xff, 0xff) # cyan
    _style = Generic.style.Style(u'Dashed Cyan Line', _lt, _color, 1.0)
    Generic.globals.styles[_style] = _style
    #
    # define and set a construction line style
    #
    _color = Generic.color.get_color(0xff, 0, 0) # red
    _lt = Generic.linetype.Linetype(u'Construction Lines', [2, 2])
    _style = Generic.style.Style(u'Construction Objects', _lt, _color, 0.0)
    Generic.conobject.ConstructionObject.__defstyle = _style
    #
    # define and add the default text style and use values in the
    # text style to define various global key/value pairs
    #
    _white = Generic.color.get_color(0xff, 0xff, 0xff)
    _ts = Generic.text.TextStyle(u'Default Text Style', color=_white)
    Generic.globals.prefs['TEXT_STYLE'] = _ts
    Generic.globals.prefs['FONT_COLOR'] = _ts.getColor()
    Generic.globals.prefs['FONT_SIZE'] = _ts.getSize()
    Generic.globals.prefs['FONT_WEIGHT'] = _ts.getWeight()
    Generic.globals.prefs['FONT_STYLE'] = _ts.getStyle()
    Generic.globals.prefs['FONT_FAMILY'] = _ts.getFamily()
    #
    # define and add the default dimension style and use the
    # values in that style to define various global preference
    # key/value pairs
    #
    _dsdict = {}
    _dsdict['DIM_PRIMARY_FONT_COLOR'] = _white
    _dsdict['DIM_SECONDARY_FONT_COLOR'] = _white
    _dimcolor = Generic.color.get_color(255, 165, 0) # orangeish
    _dsdict['DIM_COLOR'] = _dimcolor
    _ds = Generic.dimension.DimStyle(u'Default DimStyle', _dsdict)
    Generic.globals.dimstyles.append(_ds)
    Generic.globals.prefs['DIM_STYLE'] = _ds
    for _key in _ds.getOptions():
        _value = _ds.getOption(_key)
        Generic.globals.prefs[_key] = _value
    
    
def _initialize_linetypes():
    _lt = Generic.linetype.Linetype(u'Solid') # solid
    Generic.globals.linetypes[_lt] = _lt
    _lt = Generic.linetype.Linetype(u'Dash1', [4, 1]) # dashed line
    Generic.globals.linetypes[_lt] = _lt
    _lt = Generic.linetype.Linetype(u'Dash2', [8, 2]) # dashed line
    Generic.globals.linetypes[_lt] = _lt
    _lt = Generic.linetype.Linetype(u'Dash3', [12, 2]) # dashed line
    Generic.globals.linetypes[_lt] = _lt
    _lt = Generic.linetype.Linetype(u'Dash4', [10, 2, 2, 2]) # dashed line
    Generic.globals.linetypes[_lt] = _lt
    _lt = Generic.linetype.Linetype(u'Dash5', [15, 5, 5, 5]) # dashed line
    Generic.globals.linetypes[_lt] = _lt

def _initialize_colors():
    _color = Generic.color.Color(255, 0, 0) # red
    Generic.globals.colors[_color] = _color
    _color = Generic.color.Color(0, 255, 0) # green
    Generic.globals.colors[_color] = _color
    _color = Generic.color.Color(0, 0, 255) # blue
    Generic.globals.colors[_color] = _color
    _color = Generic.color.Color(255, 0, 255) # violet
    Generic.globals.colors[_color] = _color
    _color = Generic.color.Color(255, 255, 0) # yellow
    Generic.globals.colors[_color] = _color
    _color = Generic.color.Color(0, 255, 255) # cyan
    Generic.globals.colors[_color] = _color
    _color = Generic.color.Color(255, 255, 255) # white
    Generic.globals.colors[_color] = _color
    _color = Generic.color.Color(0, 0, 0) # black
    Generic.globals.colors[_color] = _color

def _initialize_globals():
    _initialize_colors()
    _initialize_linetypes()
    _initialize_styles()
    _initialize_image_colors()
    _initialize_sizes()
    _initialize_booleans()
    
def main():
    #
    # load up global and user preferences
    #

    _initialize_globals()
    Generic.preferences.initialize_prefs()
    Generic.preferences.load_global_prefs()
    _user_flag = Generic.globals.prefs['USER_PREFS']
    if _user_flag:
        Generic.preferences.load_user_prefs()
    Generic.globals.prefs.lock() # the globals are now set


if __name__ == '__main__':
    main()
    AppHelper.runEventLoop()

# need some version checking for OS X - 10.2?  10.3?
#    _major, _minor, _micro = gtk.pygtk_version
#    if ((_major < 2) and
#        (_minor < 100) and #
#        (_micro < 16)):
#        print     main()



#
# Copyright (c) 2002, 2003 Art Haas
#
# This file is part of PythonCAD.
#
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# code for chamfer and fillet objects
#

from math import hypot, pi, sin, cos, tan, atan2

import baseobject
import graphicobject
import intersections
import segment
import style
import linetype
import color
import util

_dtr = 180.0/pi

class SegJoint(graphicobject.GraphicObject):
    """A base class for chamfers and fillets

A SegJoint object has the following methods:

validate(): Check the two segments can intersect.
getSegments(): Get the two segments joined by the SegJoint object.
getMovingPoints(): Get the segment points used by the SegJoint object.
getFixedPoints(): Get the segment points not used by the SegJoint object.
update(): Recheck the SegJoint's validity.
getIntersection(): Get the intersection point of the joined segments.
inRegion(): Determine if a SegJoint is located in some area.
    """

    #
    # The default style for the Segjoint class
    #

    _defstyle = style.Style(u'Segjoint Default Style',
                            linetype.Linetype(u'Solid', None),
                            color.Color(0xffffff),
                            1.0)

    def __init__(self, s1, s2, st=None, lt=None, col=None, t=None, **kw):
        if not isinstance(s1, segment.Segment):
            raise TypeError, "Invalid first Segment for SegJoint: " + `s1`
        if not isinstance(s2, segment.Segment):
            raise TypeError, "Invalid second Segment for SegJoint: " + `s2`
        _st = st
        if _st is None:
            _st = SegJoint._defstyle
        graphicobject.GraphicObject.__init__(self, _st, lt, col, t, **kw)
        self.__s1 = s1
        self.__s2 = s2
        self.__xi = None # segment intersection x-coordinate
        self.__yi = None # segment intersection y-coordinate
        self.__s1_float = None # s1 endpoint at joint
        self.__s1_fixed = None # s1 other endpoint
        self.__s2_float = None # s2 endpoint at joint
        self.__s2_fixed = None # s2 other endpoint
        SegJoint.validate(self)
        s1.storeUser(self)
        # s1.connect('moved', self, self.moveSegment)
        s2.storeUser(self)
        # s1.connect('moved', self, self.moveSegment)

    def validate(self):
        """Check that the two segments can intersect.

validate()
        """
        _p1, _p2 = self.__s1.getEndpoints()
        _p3, _p4 = self.__s2.getEndpoints()
        if _p1 is _p3 or _p2 is _p3 or _p1 is _p4 or _p2 is _p4:
            raise ValueError, "Shared segment endpoints in s1 and s2"
        _denom = intersections.denom(_p1, _p2, _p3, _p4)
        # print "d: %g" % _denom
        if abs(_denom) < 1e-10: # parallel
            raise ValueError, "Segments are parallel"
        _rn = intersections.rnum(_p1, _p2, _p3, _p4)
        # print "rn: %g" % _rn
        _sn = intersections.snum(_p1, _p2, _p3, _p4)
        # print "sn: %g" % _sn
        _r = _rn/_denom
        _s = _sn/_denom
        if 0.0 < _r < 1.0 or 0.0 < _s < 1.0:
            raise ValueError, "Invalid segment intersection point"
        _x1, _y1 = _p1.getCoords()
        _x2, _y2 = _p2.getCoords()
        self.__xi = _x1 + _r * (_x2 - _x1) # intersection "x" coordinate
        self.__yi = _y1 + _r * (_y2 - _y1) # intersection "y" coordinate
        # print "xi: %g; yi: %g" % (self.__xi, self.__yi)
        if _r < 1e-10:
            self.__s1_fixed = _p2
            self.__s1_float = _p1
        else:
            self.__s1_fixed = _p1
            self.__s1_float = _p2
        if _s < 1e-10:
            self.__s2_fixed = _p4
            self.__s2_float = _p3
        else:
            self.__s2_fixed = _p3
            self.__s2_float = _p4

    def getSegments(self):
        """Return the two segments joined by the SegJoint.

getSegments()

This method returns a tuple holding the two segments joined
by the SegJoint.
        """
        return self.__s1, self.__s2

    def getMovingPoints(self):
        """Return the joined segment points used by the SegJoint.

getMovingPoints()

This method returns a tuple of two points, the first point is the
used point on the SegJoint initial segment, and the second point
is the used point on the SegJoint secondary segment.
        """
        return self.__s1_float, self.__s2_float

    def getFixedPoints(self):
        """Return the joined segment points not used by the SegJoint.

getFixedPoints()

This method returns a tuple of two points, the first point is the
unused point on the SegJoint initial segment, and the second point
is the unused point on the SegJoint secondary segment.
        """
        return self.__s1_fixed, self.__s2_fixed

    def update(self):
        """Revalidate the SegJoint if it is modified.

update()
        """
        if self.isModified():
            self.validate()
            self.reset()

    def getIntersection(self):
        """Return the intersection points of the SegJoint segments.

getIntersection()

This method returns a tuple of two floats; the first is the
intersection "x" coordinate, and the second is the "y" coordinate.
        """
        self.update()
        return self.__xi, self.__yi

    def inRegion(self, xmin, ymin, xmax, ymax, fully=False):
        """Return whether or not a segjoint exists with a region.

isRegion(xmin, ymin, xmax, ymax)

The four arguments define the boundary of an area, and the
function returns True if the joint lies within that area.
Otherwise, the function returns False.
        """
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        if fully is not True and fully is not False:
            raise ValueError, "Invalid flag: " + `fully`
        _mp1, _mp2 = self.getMovingPoints()
        _mx1, _my1 = _mp1.getCoords()
        _mx2, _my2 = _mp2.getCoords()
        _fxmin = min(_mx1, _mx2)
        _fymin = min(_my1, _my2)
        _fxmax = max(_mx1, _mx2)
        _fymax = max(_my1, _my2)
        if ((_fxmax < _xmin) or
            (_fymax < _ymin) or
            (_fxmin > _xmax) or
            (_fymin > _ymax)):
            return False
        if fully:
            if ((_fxmin > _xmin) and
                (_fymin > _ymin) and
                (_fxmax < _xmax) and
                (_fymax < _ymax)):
                return True
            return False
        return util.in_region(_mx1, _my1, _mx2, _my2,
                              _xmin, _ymin, _xmax, _ymax)

    def moveSegment(self, s, *args):
        _alen = len(args)
        if _alen < 4:
            raise ValueError, "Invalid argument count: %d" % _alen
        _x1 = args[0]
        if not isinstance(_x1, float):
            _x1 = float(args[0])
        _y1 = args[1]
        if not isinstance(_y1, float):
            _y1 = float(args[1])
        _x2 = args[2]
        if not isinstance(_x2, float):
            _x2 = float(args[2])
        _y2 = args[3]
        if not isinstance(_y2, float):
            _y2 = float(args[3])
        #
        # would it be better to resize the joint or to remove it?
        # we pass for now ...
        #
        if s is self.__s1:
            pass
        elif s is self.__s2:
            pass
        else:
            raise ValueError, "Unexpected segment in moveSegment" + `s`
        self.modified()
        
class Chamfer(SegJoint):
    """A Chamfer class

A chamfer is a small distance taken off a sharp
corner in a drawing. For the chamfer to be valid,
the chamfer length must be less than the length of
either segment, and the two segments must be extendable
so they could share a common endpoint.

A Chamfer is derived from a SegJoint, so it shares all
the methods and attributes of that class. A Chamfer has
the following additional methods:

{set/get}Length(): Set/Get the Chamfer length.

A Chamfer has the following attributes:

length: The Chamfer length.
    """

    messages = {
        'length_changed' : True,
        'moved' : True
        }
    
    def __init__(self, s1, s2, l, st=None, lt=None, col=None, t=None, **kw):
        SegJoint.__init__(self, s1, s2, st, lt, col, t, **kw)
        _len = l
        if not isinstance(_len, float):
            _len = float(l)
        if _len < 0.0:
            raise ValueError, "Invalid chamfer length: %g" % _len
        if _len > s1.length():
            raise ValueError, "Chamfer is longer than first Segment."
        if _len > s2.length():
            raise ValueError, "Chamfer is longer than second Segment."
        _xi, _yi = SegJoint.getIntersection(self)
        # print "xi: %g; yi: %g" % (_xi, _yi)
        _sp1, _sp2 = SegJoint.getMovingPoints(self)
        _xp, _yp = _sp1.getCoords()
        _sep = hypot((_yp - _yi), (_xp - _xi))
        if _sep > (_len + 1e-10):
            # print "sep: %g" % _sep
            # print "xp: %g; yp: %g" % (_xp, _yp)
            raise ValueError, "First segment too far from intersection point."
        _xp, _yp = _sp2.getCoords()
        _sep = hypot((_yp - _yi), (_xp - _xi))
        if _sep > (_len + 1e-10):
            # print "sep: %g" % _sep
            # print "xp: %g; yp: %g" % (_xp, _yp)
            raise ValueError, "Second segment too far from intersection point."
        self.__length = _len
        self._moveSegmentPoints(_len)

    def __eq__(self, obj):
        if not isinstance(obj, Chamfer):
            return False
        _s1, _s2 = self.getSegments()
        _os1, _os2 = obj.getSegments()
        _val = False
        if ((_s1 == _os1 and _s2 == _os2) or
            (_s1 == _os2 and _s2 == _os1)):
            if abs(self.__length - obj.getLength()) < 1e-10:
                _val = True
        return _val

    def __ne__(self, obj):
        if not isinstance(obj, Chamfer):
            return True
        _s1, _s2 = self.getSegments()
        _os1, _os2 = obj.getSegments()
        _val = True
        if ((_s1 == _os1 and _s2 == _os2) or
            (_s1 == _os2 and _s2 == _os1)):
            if abs(self.__length - obj.getLength()) < 1e-10:
                _val = False
        return _val

    def getValues(self):
        _pids = SegJoint.getValues(self)
        _gv = self.getGraphicValues(SegJoint._defstyle)
        _s1, _s2 = self.getSegments()
        return 'chamfer', _pids, _gv, _s1.getID(), _s2.getID(), self.__length

    def getLength(self):
        """Return the Chamfer length.

getLength()
        """
        return self.__length

    def setLength(self, l):
        """Set the Chamfer length.

setLength(l)

The length should be a positive float value.
        """
        _s1, _s2 = self.getSegments()        
        if (self.isLocked() or
            _s1.isLocked() or
            _s2.isLocked()):
            raise RuntimeError, "Setting length not allowed - object locked."
        _l = l
        if not isinstance(_l, float):
            _l = float(l)
        if _l < 0.0:
            raise ValueError, "Invalid chamfer length: %g" % _l
        if _l > _s1.length():
            raise ValueError, "Chamfer is larger than first Segment."
        if _l > _s2.length():
            raise ValueError, "Chamfer is larger than second Segment."
        _ol = self.__length
        if abs(_l - _ol) > 1e-10:
            self.__length = _l
            self.sendMessage('length_changed', _ol)
            self._moveSegmentPoints(_l)
            self.modified()

    length = property(getLength, setLength, None, "Chamfer length.")

    def _moveSegmentPoints(self, dist):
        """Set the Chamfer endpoints at the correct location

moveSegmentPoints(dist)

The argument "dist" is the chamfer length. This method is private
the the Chamfer object.
        """
        _d = dist
        if not isinstance(dist, float):
            _d = float(dist)
        #
        # process segment 1
        #
        _xi, _yi = self.getIntersection()
        # print "xi: %g; yi: %g" % (xi, yi)
        _mp1, _mp2 = self.getMovingPoints()
        _sp1, _sp2 = self.getFixedPoints()
        _sx, _sy = _sp1.getCoords()
        _slen = hypot((_yi - _sy), (_xi - _sx))
        # print "slen: %g" % slen
        _newlen = (_slen - _d)/_slen
        # print "newlen: %g" % _newlen
        _xs, _ys = _sp1.getCoords()
        _xm, _ym = _mp1.getCoords()
        _xn = _xs + _newlen * (_xi - _xs)
        _yn = _ys + _newlen * (_yi - _ys)
        # print "xn: %g; yn: %g" % (_xn, _yn)
        _mp1.x = _xn
        _mp1.y = _yn
        #
        # process segment 2
        #
        _sx, _sy = _sp2.getCoords()
        _slen = hypot((_yi - _sy), (_xi - _sx))
        # print "slen: %g" % _slen
        _newlen = (_slen - _d)/_slen
        # print "newlen: %g" % _newlen
        _xs, _ys = _sp2.getCoords()
        _xm, _ym = _mp2.getCoords()
        _xn = _xs + _newlen * (_xi - _xs)
        _yn = _ys + _newlen * (_yi - _ys)
        # print "xn: %g; yn: %g" % (_xn, _yn)
        _mp2.x = _xn
        _mp2.y = _yn

    def sendsMessage(self, m):
        if m in Chamfer.messages:
            return True
        return SegJoint.sendsMessage(self, m)

#
# Chamfer history class
#

class ChamferLog(graphicobject.GraphicObjectLog):
    def __init__(self, c):
        if not isinstance(c, Chamfer):
            raise TypeError, "Invalid chamfer: " + `c`
        graphicobject.GraphicObjectLog.__init__(self, c)
        c.connect('length_changed', self, ChamferLog.lengthChange)

    def lengthChange(self, c, *args):
        _alen = len(args)
        if _alen < 1:
            raise ValueError, "Invalid argument count: %d" % _alen
        _l = args[0]
        if not isinstance(_l, float):
            _l = float(args[0])
        self.saveUndoData('length_changed', _l)

    def execute(self, undo, *args):
        if undo is not True and undo is not False:
            raise ValueError, "Invalid undo value: " + str(undo)
        _alen = len(args)
        if _alen == 0:
            raise ValueError, "No arguments to execute()"
        _c = self.getObject()
        _op = args[0]
        if _op == 'length_changed':
            if len(args) < 2:
                raise ValueError, "Invalid argument count: %d" % _alen
            _l = args[1]
            if not isinstance(_l, float):
                _l = float(args[1])
            _cl = _c.getLength()
            if undo:
                self.saveUndoData('length_changed', _cl)
            else:
                self.saveRedoData('length_changed', _cl)
            self.ignore('length_changed')
            try:
                if undo:
                    _moved = False
                    _c.startUndo(True)
                    try:
                        _c.setLength(_l)
                        if abs(_l - _cl) > 1e-10:
                            _moved = True
                    finally:
                        _c.endUndo()
                    if _moved:
                        pass # self.sendMessage('moved', _cl)
                else:
                    _c.setLength(_l)
            finally:
                self.receive('length_changed')
        else:
            raise ValueError, "Unexpected operation: %s" % _op

class Fillet(SegJoint):
    """A Fillet class

A fillet is a curved joining of two segments. For a filleted
joint to be valid, the radius must fall within some distance
determined by the segment endpoints and segment intersection
point, and the two segments must be extendable so they can
share a common endpoint.

A Fillet is derived from a SegJoint, so it shares the methods
and attributes of that class. A Fillet has the following additional
methods:

{set/get}Radius(): Set/Get the fillet radius.
getCenter(): Get the center point of the fillet.
getAngles(): Get the angles the fillet sweeps between.
    """
    messages = {
    'radius_changed' : True,
    'moved' : True
    }
    def __init__(self, s1, s2, r, st=None, lt=None, col=None, t=None, **kw):
        SegJoint.__init__(self, s1, s2, st, lt, col, t, **kw)
        _r = r
        if not isinstance(_r, float):
            _r = float(r)
        if _r < 0.0:
            raise ValueError, "Invalid fillet radius: %g" % _r
        self._calculateLimits()
        _rmin, _rmax = self.getRadialLimits()
        if _r < _rmin or _r > _rmax:
            raise ValueError, "Invalid radius: %g" % _r
        self.__radius = _r
        self.__center = (0.0, 0.0)
        self._calculateCenter()
        self._moveSegmentPoints()

    def __eq__(self, obj):
        if not isinstance(obj, Fillet):
            return False
        _s1, _s2 = self.getSegments()
        _os1, _os2 = obj.getSegments()
        _val = False
        if ((_s1 == _os1 and _s2 == _os2) or
            (_s1 == _os2 and _s2 == _os1)):
            if abs(self.__radius - obj.getRadius()) < 1e-10:
                _val = True
        return _val

    def __ne__(self, obj):
        if not isinstance(obj, Fillet):
            return True
        _s1, _s2 = self.getSegments()
        _os1, _os2 = obj.getSegments()
        _val = True
        if ((_s1 == _os1 and _s2 == _os2) or
            (_s1 == _os2 and _s2 == _os1)):
            if abs(self.__radius - obj.getRadius()) < 1e-10:
                _val = False
        return _val

    def getValues(self):
        _pids = SegJoint.getValues(self)
        _gv = self.getGraphicValues(SegJoint._defstyle)
        _s1, _s2 = self.getSegments()
        return 'fillet', _pids, _gv, _s1.getID(), _s2.getID(), self.__radius

    def getRadius(self):
        """Return the Fillet radius.

getRadius()
        """
        return self.__radius

    def setRadius(self, r):
        """Set the Fillet radius.

setRadius(r)

The radius should be a positive float value.
        """
        _s1, _s2 = self.getSegments()        
        if (self.isLocked() or
            _s1.isLocked() or
            _s2.isLocked()):
            raise RuntimeError, "Setting length not allowed - object locked."
        _r = r
        if not isinstance(_r, float):
            _r = float(r)
        if _r < 0.0:
            raise ValueError, "Invalid fillet radius: %g" % _r
        self._calculateLimits()
        _rmin, _rmax = self.getRadialLimits()
        if _r < _rmin or _r > _rmax:
            raise ValueError, "Invalid radius: %g" % _r
        _or = self.__radius
        if abs(_r - _or) > 1e-10:
            self.__radius = _r
            self._calculateCenter()
            self._moveSegmentPoints()
            self.sendMessage('radius_changed', _or)

    radius = property(getRadius, setRadius, None, "Chamfer radius.")

    def _calculateCenter(self):
        """Find the center point of the radius

_calculateCenter()

This method is private to the Fillet object.
        """
        _r = self.__radius
        _p1, _p3 = self.getMovingPoints()
        _p2, _p4 = self.getFixedPoints()
        _as1 = atan2((_p2.y - _p1.y), (_p2.x - _p1.x)) # _as1 in radians
        _as2 = atan2((_p4.y - _p3.y), (_p4.x - _p3.x)) # _as2 in radians
        if abs(abs(_as1) - pi) < 1e-10:
            if _as1 > 0.0 and _as2 < 0.0:
                _as1 = -pi
            if _as1 < 0.0 and _as2 > 0.0:
                _as1 = pi
        if abs(abs(_as2) - pi) < 1e-10:
            if _as2 > 0.0 and _as2 < 0.0:
                _as2 = -pi
            if _as2 < 0.0 and _as1 > 0.0:
                _as2 = pi
        _acl = (_as1 + _as2)/2.0
        _acc = abs(_as1 - _as2)/2.0
        if (_as1 > 0.0 and _as2 < 0.0) or (_as1 < 0.0 and _as2 > 0.0):
            _amin = min(_as1, _as2)
            _amax = max(_as1, _as2)
            print "_amax: %g" % _amax
            print "_amin: %g" % _amin
            if _amax - _amin > pi: # radians
                if _acl < 0.0:
                    _acl = _acl + pi
                else:
                    _acl = _acl - pi
                _acc = ((pi - _amax) + (_amin + pi))/2.0
        print "_acl: %g" % (_acl * _dtr)
        print "_acc: %g" % (_acc * _dtr)
        _rc = hypot((_r/tan(_acc)), _r)
        print "_rc: %g" % _rc
        _xi, _yi = self.getIntersection()
        _xc = _xi + _rc * cos(_acl)
        _yc = _yi + _rc * sin(_acl)
        self.__center = (_xc, _yc)
        print "center: %s" % str(self.__center)

    def getCenter(self):
        """Return the center location of the Fillet.

getCenter()

This method returns a tuple of two floats; the first is the
center "x" coordinate, the second is the "y" coordinate.
        """
        return self.__center

    def _calculateLimits(self):
        """Determine the radial limits of the fillet.

_calculateLimits()

This method is private to the Fillet.
        """
        _p1, _p3 = self.getMovingPoints()
        _p2, _p4 = self.getFixedPoints()
        _as1 = atan2((_p2.y - _p1.y), (_p2.x - _p1.x)) # radians
        _as2 = atan2((_p4.y - _p3.y), (_p4.x - _p3.x)) # radians
        if abs(abs(_as1) - pi) < 1e-10:
            if _as1 > 0.0 and _as2 < 0.0:
                _as1 = -pi
            if _as1 < 0.0 and _as2 > 0.0:
                _as1 = pi
        if abs(abs(_as2) - pi) < 1e-10:
            if _as2 > 0.0 and _as2 < 0.0:
                _as2 = -pi
            if _as2 < 0.0 and _as1 > 0.0:
                _as2 = pi
        print "_as1: %g" % (_as1 * _dtr)
        print "_as2: %g" % (_as2 * _dtr)
        _acl = (_as1 + _as2)/2.0
        _acc = abs(_as1 - _as2)/2.0
        if (_as1 > 0.0 and _as2 < 0.0) or (_as1 < 0.0 and _as2 > 0.0):
            _amin = min(_as1, _as2)
            _amax = max(_as1, _as2)
            print "_amax: %g" % _amax
            print "_amin: %g" % _amin
            if _amax - _amin > pi: # radians
                if _acl < 0.0:
                    _acl = _acl + pi
                else:
                    _acl = _acl - pi
                _acc = ((pi - _amax) + (_amin + pi))/2.0
        print "_acl: %g" % (_acl * _dtr)
        print "_acc: %g" % (_acc * _dtr)
        _xi, _yi = self.getIntersection()
        _pf1, _pf2 = self.getFixedPoints()
        _d1 = hypot((_xi - _pf1.x), (_yi - _pf1.y))
        _d2 = hypot((_xi - _pf2.x), (_yi - _pf2.y))
        _c4 = min(_d1, _d2)
        self.__rmax = _c4 * tan(_acc) + 1e-10
        print "rmax: %g" % self.__rmax
        _pm1, _pm2 = self.getMovingPoints()
        _d1 = hypot((_xi - _pm1.x), (_yi - _pm1.y))
        _d2 = hypot((_xi - _pm2.x), (_yi - _pm2.y))
        _c4 = max(_d1, _d2)
        self.__rmin = _c4 * tan(_acc) - 1e-10
        print "rmin: %g" % self.__rmin

    def getRadialLimits(self):
        """Return the radial limits of the fillet.

getRadialLimits()

This method returns a tuple of two floats; the first is
the minimal radius for the fillet between two segments,
and the second is the maximum radius.
        """
        return self.__rmin, self.__rmax

    def _moveSegmentPoints(self):
        """Position the segment endpoints used in the Fillet.

_moveSegmentPoints()

This method is private to the Fillet.
        """
        _p1, _p3 = self.getMovingPoints()
        _p2, _p4 = self.getFixedPoints()
        _xc, _yc = self.__center
        #
        # segment 1
        #
        _l = _p2 - _p1
        _x1, _y1 = _p1.getCoords()
        _x2, _y2 = _p2.getCoords()
        _r = ((_xc - _x1)*(_x2 - _x1) + (_yc - _y1)*(_y2 - _y1))/pow(_l, 2)
        _px = _x1 + _r * (_x2 - _x1)
        _py = _y1 + _r * (_y2 - _y1)
        _p1.x = _px
        _p1.y = _py
        #
        # segment 2
        #
        _l = _p4 - _p3
        _x1, _y1 = _p3.getCoords()
        _x2, _y2 = _p4.getCoords()
        _r = ((_xc - _x1)*(_x2 - _x1) + (_yc - _y1)*(_y2 - _y1))/pow(_l, 2)
        _px = _x1 + _r * (_x2 - _x1)
        _py = _y1 + _r * (_y2 - _y1)
        _p3.x = _px
        _p3.y = _py

    def getAngles(self):
        """Return the angles that the fillet sweeps through.

getAngles()

This method returns a tuple of two floats, the first is the
start angle of the fillet, and the second is the end angle.
        """
        _ms1, _ms2 = self.getMovingPoints()
        _xc, _yc = self.__center
        _x, _y = _ms1.getCoords()
        _as1 = _dtr * atan2((_y - _yc), (_x - _xc))
        if _as1 < 0.0:
            _as1 = _as1 + 360.0
        _x, _y = _ms2.getCoords()
        _as2 = _dtr * atan2((_y - _yc), (_x - _xc))
        if _as2 < 0.0:
            _as2 = _as2 + 360.0
        return _as1, _as2

    def inRegion(self, xmin, ymin, xmax, ymax, fully=False):
        """Return whether or not a fillet exists with a region.

isRegion(xmin, ymin, xmax, ymax)

The four arguments define the boundary of an area, and the
function returns True if the joint lies within that area.
Otherwise, the function returns False.
        """
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _xmax = xmax
        if not isinstance(_xmax, float):
            _xmax = float(xmax)
        if _xmax < _xmin:
            raise ValueError, "Illegal values: xmax < xmin"
        _ymax = ymax
        if not isinstance(_ymax, float):
            _ymax = float(ymax)
        if _ymax < _ymin:
            raise ValueError, "Illegal values: ymax < ymin"
        if fully is not True and fully is not False:
            raise ValueError, "Invalid flag: " + `fully`
        _mp1, _mp2 = self.getMovingPoints()
        _mx1, _my1 = _mp1.getCoords()
        _mx2, _my2 = _mp2.getCoords()
        _r = self.__radius
        _xc, _yc = self.__center
        _a1, _a2 = self.getAngles()
        _xl = [_mx1, _mx2, _xc]
        _yl = [_my1, _my2, _yc]
        if fully:
            if ((min(_xl) > _xmin) and
                (min(_yl) > _ymin) and
                (max(_xl) < _xmax) and
                (max(_yl) < _ymax)):
                return True
            return False
        #
        # fixme - need to use the arc and endpoints and not
        # a line connecting the endpoints ...
        #
        return util.in_region(_mx1, _my1, _mx2, _my2,
                              _xmin, _ymin, _xmax, _ymax)

    def sendsMessage(self, m):
        if m in Fillet.messages:
            return True
        return SegJoint.sendsMessage(self, m)

#
# Fillet history class
#

class FilletLog(graphicobject.GraphicObjectLog):
    def __init__(self, f):
        if not isinstance(f, Fillet):
            raise TypeError, "Invalid fillet: " + `f`
        graphicobject.GraphicObjectLog.__init__(self, f)
        f.connect('radius_changed', self, FilletLog.radiusChange)

    def radiusChange(self, f, *args):
        _alen = len(args)
        if _alen < 1:
            raise ValueError, "Invalid argument count: %d" % _alen
        _r = args[0]
        if not isinstance(_r, float):
            _r = float(args[0])
        self.saveUndoData('radius_changed', _r)

    def execute(self, undo, *args):
        if undo is not True and undo is not False:
            raise ValueError, "Invalid undo value: " + str(undo)
        _alen = len(args)
        if _alen == 0:
            raise ValueError, "No arguments to execute()"
        _f = self.getObject()
        _op = args[0]
        if _op == 'radius_changed':
            if len(args) < 2:
                raise ValueError, "Invalid argument count: %d" % _alen
            _r = args[1]
            if not isinstance(_r, float):
                _r = float(args[1])
            _cr = _f.getRadius()
            if undo:
                self.saveUndoData('radius_changed', _cr)
            else:
                self.saveRedoData('radius_changed', _cr)
            self.ignore('radius_changed')
            try:
                if undo:
                    _moved = False
                    _f.startUndo(True)
                    try:
                        _f.setRadius(_r)
                        if abs(_r - _cr) > 1e-10:
                            _moved = True
                    finally:
                        _f.endUndo()
                    if _moved:
                        pass # _f.sendMessage('moved', _cr)
                else:
                    _f.setRadius(_r)
            finally:
                self.receive('radius_changed')
        else:
            raise ValueError, "Unexpected operation: %s" % _op



#!/usr/bin/python
#
# simply binary dump
#

import sys
import array

if len(sys.argv) < 4:
    print "bindump.py file offset count"
    sys.exit(1)
else:
    try:
        _fname = file(sys.argv[1])
    except:
        print "invalid file: " + sys.argv[1]
        sys.exit(1)
    try:
        _offset = int(sys.argv[2])
    except:
        print "invalid offset: " + sys.argv[2]
        sys.exit(1)
    if _offset < 0:
        print "invalid offset: %d" % _offset
        sys.exit(1)
    try:
        _count = int(sys.argv[3])
    except:
        print "invalid byte count: " + sys.argv[3]
        sys.exit(1)
    if _count < 0:
        print "invalid byte count: %d" % _count
        sys.exit(1)

print "opening file: " + _fname.name
print "offset: %d" % _offset
print "count: %d" % _count 

try:
    _fname.seek(_offset, 0)
except:
    _fname.close()
    print "invalid offset into file: %d" % _offset
    sys.exit(1)

_data = array.array('B')
try:
    _data.fromfile(_fname, _count)
except:
    _fname.close()
    print "invalid read of %d bytes from file: %s" % (_count, _fname.name)
    sys.exit(1)

_fname.close()

_patterns = [
    '0 0 0 0', # 0
    '0 0 0 1', # 1
    '0 0 1 0', # 2
    '0 0 1 1', # 3
    '0 1 0 0', # 4
    '0 1 0 1', # 5
    '0 1 1 0', # 6
    '0 1 1 1', # 7
    '1 0 0 0', # 8
    '1 0 0 1', # 9
    '1 0 1 0', # A
    '1 0 1 1', # B
    '1 1 0 0', # C
    '1 1 0 1', # D
    '1 1 1 0', # E
    '1 1 1 1'  # F
    ]

_i = 0
while (_i < _count):
    _bitoffset = _i * 8
    _nib1 = _patterns[((_data[_i] & 0xf0) >> 4)]
    _nib2 = _patterns[(_data[_i] & 0x0f)]
    print "%d [%d]: 0x%02x %s %s" % (_i, _bitoffset, _data[_i], _nib1, _nib2)
    _i = _i + 1



#
# Copyright (c) 2002, 2003, 2004 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# available units
#

MILLIMETERS = 0
MICROMETERS = 1
METERS = 2
KILOMETERS = 3
INCHES = 4
FEET = 5
YARDS = 6
MILES = 7

def get_all_units():
    unitlist = []
    unitlist.append('Millimeters')
    unitlist.append('Micrometers')
    unitlist.append('Meters')
    unitlist.append('Kilometers')
    unitlist.append('Inches')
    unitlist.append('Feet')
    unitlist.append('Yards')
    unitlist.append('Miles')
    return unitlist

class Unit(object):
    """A class defining a unit for linear dimensions.

The Unit class is used to establish what unit to
assign distances between points. The class contains
two methods used to scale distance value to and from
an equivalent distance in millimeters.

toMillimeters(): Convert a distance to millimeters
fromMillimeters(): Convert a distance from millimeters
    """
    def __init__(self, unit=MILLIMETERS):
        if (unit != MILLIMETERS and
            unit != MICROMETERS and
            unit != METERS and
            unit != KILOMETERS and
            unit != INCHES and
            unit != FEET and
            unit != YARDS and
            unit != MILES):
            raise ValueError, "Invalid unit choice: " + str(unit)
        self.__unit = unit

    def getUnit(self):
        return self.__unit

    def getStringUnit(self):
        _u = self.__unit
        if _u == MILLIMETERS:
            _str = 'millimeters'
        elif _u == MICROMETERS:
            _str = 'micrometers'
        elif _u == METERS:
            _str = 'meters'
        elif _u == KILOMETERS:
            _str = 'kilometers'
        elif _u == INCHES:
            _str = 'inches'
        elif _u == FEET:
            _str = 'feet'
        elif _u == YARDS:
            _str = 'yards'
        elif _u == MILES:
            _str = 'miles'
        else:
            raise ValueError, "Unexpected unit: %d" % _u
        return _str
    
    def setUnit(self, unit):
        if (unit != MILLIMETERS and
            unit != MICROMETERS and
            unit != METERS and
            unit != KILOMETERS and
            unit != INCHES and
            unit != FEET and
            unit != YARDS and
            unit != MILES):
            raise ValueError, "Invalid unit choice: " + str(unit)
        self.__unit = unit

    unit = property(getUnit, setUnit, None, "Basic unit.")

    def toMillimeters(self, value):
        """Scale a value to the equivalent distance in millimeters.

toMillimeters(value)
        """
        _v = value
        if not isinstance(_v, float):
            _v = float(value)
        if self.__unit == MILLIMETERS:
            _sv = _v
        elif self.__unit == MICROMETERS:
            _sv = _v * 1e-3
        elif self.__unit == METERS:
            _sv = _v * 1e3
        elif self.__unit == KILOMETERS:
            _sv = _v * 1e6
        elif self.__unit == INCHES:
            _sv = _v * 25.4
        elif self.__unit == FEET:
            _sv = _v * 304.8
        elif self.__unit == YARDS:
            _sv = _v * 914.4
        elif self.__unit == MILES:
            _sv = _v * 1609344.4
        else:
            raise ValueError, "Undefined unit value! " + str(self.__unit)
        return _sv

    def fromMillimeters(self, value):
        """Scale a value from an equivalent distance in millimeters.

fromMillimeters(value)
        """
        _v = value
        if not isinstance(_v, float):
            _v = float(value)
        if self.__unit == MILLIMETERS:
            _sv = _v
        elif self.__unit == MICROMETERS:
            _sv = _v * 1e3
        elif self.__unit == METERS:
            _sv = _v * 1e-3
        elif self.__unit == KILOMETERS:
            _sv = _v * 1e-6
        elif self.__unit == INCHES:
            _sv = _v / 25.4
        elif self.__unit == FEET:
            _sv = _v / 304.8
        elif self.__unit == YARDS:
            _sv = _v / 914.4
        elif self.__unit == MILES:
            _sv = _v / 1609344.4
        else:
            raise ValueError, "Undefined unit value! " + str(self.__unit)
        return _sv

def unit_string(value):
    """Return a text string for the integer unit value.

unit_string(value)
    """
    if value == MILLIMETERS:
        _str = 'millimeters'
    elif value == MICROMETERS:
        _str = 'micrometers'
    elif value == METERS:
        _str = 'meters'
    elif value == KILOMETERS:
        _str = 'kilometers'
    elif value == INCHES:
        _str = 'inches'
    elif value == FEET:
        _str = 'feet'
    elif value == YARDS:
        _str = 'yards'
    elif value == MILES:
        _str = 'miles'
    else:
        raise ValueError, "Unexpected unit: " + str(value)
    return _str
        
        
